from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from uuid import uuid4

import compas_rhino
from compas_3gs.rhino.objects import VolMeshObject
from compas_3gs.rhino.objects import NetworkObject

import scriptcontext as sc

__all__ = ['Scene']


class Scene(object):
    """A Rhino scene for 3GS.

    Attributes
    ----------
    objects : dict
        Mapping between GUIDs and diagram objects added to the scene.
        The GUIDs are automatically generated and assigned.

    Examples
    --------
    .. code-block:: python

        from compas_ags.diagrams import FormGraph
        from compas_ags.diagrams import FormDiagram

        from compas_ags.rhino import Scene

        graph = FormGraph.from_obj(FILE)
        form = FormDiagram.from_graph(graph)

        scene = Scene()

        guid = scene.add(form, name="Form", layer="AGS::FormDiagram")
        form = scene.find(guid)

        scene.clear()
        scene.update()

    """

    def __init__(self, db=None, depth=10, settings=None):
        self._current = -1
        self._depth = depth
        self._db = db
        self.objects = {}
        self.settings = settings or {}

    def add_forcevolmesh(self, item, name=None, layer=None, visible=True, settings=None):
        """Add an object to the scene matching the provided item.

        Parameters
        ----------
        item : :class:`compas_ags.diagrams.Diagram`
        name : str, optional
        layer : str, optional
        visible : bool, optional
        settings : dict, optional

        Returns
        -------
        GUID
        """
        guid = uuid4()
        obj = VolMeshObject.build(item, scene=self, name=name, layer=layer, visible=visible, settings=settings)
        self.objects[guid] = obj
        return guid

    def add_formnetwork(self, item, name=None, layer=None, visible=True, settings=None):
        """Add an object to the scene matching the provided item.

        Parameters
        ----------
        item : :class:`compas_ags.diagrams.Diagram`
        name : str, optional
        layer : str, optional
        visible : bool, optional
        settings : dict, optional

        Returns
        -------
        GUID
        """
        guid = uuid4()
        obj = NetworkObject.build(item, scene=self, name=name, layer=layer, visible=visible, settings=settings)
        self.objects[guid] = obj
        return guid

    def find(self, guid):
        """Find an object using its GUID.

        Parameters
        ----------
        guid : str

        Returns
        -------
        :class:`compas_ags.rhino.DiagramObject`
        """
        if guid in self.objects:
            return self.objects[guid]

    def find_by_name(self, name):
        """Find an object using its name.

        Parameters
        ----------
        name : str

        Returns
        -------
        list of :class:`compas_ags.rhino.DiagramObject`
        """
        objects = []
        for obj in self.objects.values():
            if obj.name == name:
                objects.append(obj)
        return objects

    def purge(self):
        """Clear all objects from the scene."""
        compas_rhino.rs.EnableRedraw(False)
        try:
            for guid in list(self.objects.keys()):
                self.objects[guid].clear()
                del self.objects[guid]
        except Exception:
            pass
        compas_rhino.rs.EnableRedraw(True)
        compas_rhino.rs.Redraw()

    def clear(self):
        """Clear all objects from the scene."""
        compas_rhino.rs.EnableRedraw(False)
        try:
            for guid in list(self.objects.keys()):
                self.objects[guid].clear()
        except Exception:
            pass
        compas_rhino.rs.EnableRedraw(True)
        compas_rhino.rs.Redraw()

    def clear_layers(self):
        """Clear all object layers of the scene."""
        compas_rhino.rs.EnableRedraw(False)
        try:
            for guid in list(self.objects.keys()):
                self.objects[guid].clear_layer()
        except Exception:
            pass
        compas_rhino.rs.EnableRedraw(True)
        compas_rhino.rs.Redraw()

    def redraw(self):
        """Redraw the entire scene."""
        compas_rhino.rs.EnableRedraw(False)
        try:
            for guid in self.objects:
                self.objects[guid].draw()
        except Exception:
            pass
        compas_rhino.rs.EnableRedraw(True)
        compas_rhino.rs.Redraw()

    def update(self):
        """Redraw all objects in the scene."""
        self.clear()
        self.redraw()

    def save(self):
        if not self._db:
            return
        states = self._db['states']
        if self._current < -1:
            del states[self._current + 1:]
        self._current = -1
        state = []
        for guid, obj in self.objects.items():
            # the definition of a state should be formalised
            # this is equivalent to the data schema of data objects
            # the scene has a state schema
            # whether or not to store a state could be the responsibility of the caller...
            state.append({
                'object': {
                    'name': obj.name,
                    'layer': obj.layer,
                    'visible': obj.visible,
                    'settings': obj.settings,
                    'anchor': obj.anchor,
                    'location': list(obj.location),
                    'scale': obj.scale,
                },
                'diagram': {
                    'type': type(obj.diagram),
                    'data': obj.diagram.to_data(),
                },
            })
        states.append(state)
        if len(states) > self._depth:
            del states[0]
        self._db['states'] = states

        # Insert custom undo/redo event
        def undo_redo(sender, e):
            if e.Tag == "undo":
                print("running 3gs undo")
                if self.undo():
                    sc.doc.AddCustomUndoEvent("Custom_undo_redo", undo_redo, "redo")
            if e.Tag == "redo":
                print("running 3gs redo")
                if self.redo():
                    sc.doc.AddCustomUndoEvent("Custom_undo_redo", undo_redo, "undo")
        sc.doc.AddCustomUndoEvent("Custom_undo_redo", undo_redo, "undo")

    def undo(self):
        """Undo scene updates.

        Returns
        -------
        bool
            False if there is nothing (more) to undo.
            True if undo was successful.
        """
        if not self._db:
            return
        if self._current <= - self._depth:
            return False
        if len(self._db['states']) < 2:
            return False
        self.clear_layers()
        self.purge()
        self._current -= 1
        state = self._db['states'][self._current]
        form = None
        force = None
        for data in state:
            diagram = data['diagram']['type'].from_data(data['diagram']['data'])
            if data['object']['name'] == 'form':
                guid = self.add_formnetwork(diagram, name=data['object']['name'], layer=data['object']['layer'],
                                            visible=data['object']['visible'], settings=data['object']['settings'])
            if data['object']['name'] == 'force':
                guid = self.add_forcevolmesh(diagram, name=data['object']['name'], layer=data['object']['layer'],
                                             visible=data['object']['visible'], settings=data['object']['settings'])

            obj = self.find(guid)
            obj.anchor = data['object']['anchor']
            obj.location = data['object']['location']
            obj.scale = data['object']['scale']

            if obj.name == 'form':
                form = obj
            elif obj.name == 'force':
                force = obj

        if form and force:
            form.diagram.dual = force.diagram
            force.diagram.primal = form.diagram
        self.redraw()
        return True

    def redo(self):
        """Redo scene updates.

        Returns
        -------
        bool
            False if there is nothing (more) to redo.
            True if redo was successful.
        """
        if not self._db:
            return
        if len(self._db['states']) < 2:
            return False
        if self._current >= -1:
            return False
        self.clear_layers()
        self.purge()
        self._current += 1
        state = self._db['states'][self._current]
        form = None
        force = None
        for data in state:
            diagram = data['diagram']['type'].from_data(data['diagram']['data'])
            if data['object']['name'] == 'form':
                guid = self.add_formnetwork(diagram, name=data['object']['name'], layer=data['object']['layer'],
                                            visible=data['object']['visible'], settings=data['object']['settings'])
            if data['object']['name'] == 'force':
                guid = self.add_forcevolmesh(diagram, name=data['object']['name'], layer=data['object']['layer'],
                                             visible=data['object']['visible'], settings=data['object']['settings'])
            obj = self.find(guid)
            obj.anchor = data['object']['anchor']
            obj.location = data['object']['location']
            obj.scale = data['object']['scale']
            if obj.name == 'form':
                form = obj
            elif obj.name == 'force':
                force = obj
        if form and force:
            form.diagram.dual = force.diagram
            force.diagram.primal = form.diagram
        self.redraw()
        return True



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import errno
import shelve

import scriptcontext as sc

import compas
import compas_rhino

from compas_3gs.rhino import Scene
from compas_3gs.rhino import Browser
from compas_3gs.activate import check
from compas_3gs.activate import activate


__commandname__ = "TGS__init"


SETTINGS = {

    "3GS": {
        "show.angles": True,
        "show.forces": False,
        "tol.angles": 1.0,
        "tol.flatness": 0.1
    },

    "Solvers": {
        "reciprocation.alpha": 1.0,
        "reciprocation.l_min": 0.1,
        "reciprocation.l_max": 100000,
        "reciprocation.kmax": 500,
        "reciprocation.tol": 0.01,
        "reciprocation.refreshrate": 10,

        "planarization.kmax": 500,
        "planarization.tol": 0.01,
        "planarization.refreshrate": 10,

        "arearization.kmax": 500,
        "arearization.tol": 0.01,
        "arearization.refreshrate": 10,

    }
}


HERE = compas_rhino.get_document_dirname()
HOME = os.path.expanduser('~')
CWD = HERE or HOME


def RunCommand(is_interactive):

    if check():
        print("Current plugin is already activated")
    else:
        compas_rhino.rs.MessageBox("Detected environment change, re-activating plugin", 0, "Re-activating Needed")
        if activate():
            compas_rhino.rs.MessageBox("Restart Rhino for the change to take effect", 0, "Restart Rhino")
        else:
            compas_rhino.rs.MessageBox("Someting wrong during re-activation", 0, "Error")
        return

    Browser()

    shelvepath = os.path.join(compas.APPTEMP, '3GS', '.history')
    if not os.path.exists(os.path.dirname(shelvepath)):
        try:
            os.makedirs(os.path.dirname(shelvepath))
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise

    db = shelve.open(shelvepath, 'n')
    db['states'] = []

    scene = Scene(db, 20, SETTINGS)
    scene.purge()

    sc.sticky["3GS"] = {
        'system': {
            "session.dirname": CWD,
            "session.filename": None,
            "session.extension": '3gs'
        },
        'scene': scene,
    }

    scene.update()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

from compas.geometry import add_vectors
from compas.geometry import subtract_vectors
from compas.geometry import dot_vectors
from compas.geometry import scale_vector
from compas.geometry import normalize_vector
from compas.geometry import intersection_line_plane
from compas.geometry import distance_point_point
from compas.geometry import centroid_points
from compas.utilities import i_to_rgb

from compas_rhino.objects import mesh_select_vertices
from compas_rhino.objects import mesh_select_faces

from compas_3gs.operations import volmesh_vertex_merge
from compas_3gs.operations import volmesh_vertex_lift
from compas_3gs.operations import volmesh_halfface_pinch
from compas_3gs.operations import volmesh_merge_adjacent_halffaces
from compas_3gs.operations import volmesh_cell_subdivide_barycentric

from compas_3gs.rhino import VolMeshSelector
from compas_3gs.rhino import CellSelector
from compas_3gs.rhino import VolmeshCellInspector

from compas_3gs.rhino import get_target_point


try:
    import Rhino
    import rhinoscriptsyntax as rs
    import scriptcontext as sc

    from System.Drawing.Color import FromArgb

    from Rhino.Geometry import Point3d
    from Rhino.Geometry import Vector3d
    from Rhino.Geometry import Line

    find_object = sc.doc.Objects.Find
    feedback_color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
    arrow_color = FromArgb(255, 0, 79)
    jl_blue = FromArgb(0, 113, 188)
    black = FromArgb(0, 0, 0)
    gray = FromArgb(200, 200, 200)
    green = FromArgb(0, 255, 0)
    white = FromArgb(255, 255, 255)

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['rhino_volmesh_vertex_lift',
           'rhino_volmesh_vertex_merge',

           'rhino_volmesh_halfface_pinch',
           'rhino_volmesh_merge_adjacent_halffaces',

           'rhino_volmesh_cell_subdivide_barycentric',
           'rhino_volmesh_pull_halffaces']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   vertex operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def rhino_volmesh_vertex_lift(volmesh):
    """Rhino wrapper for the vertex lift operation.
    """

    vertex = VolMeshSelector.select_vertex(volmesh)

    vertex_hfkeys = []
    for hfkey in volmesh.vertex_halffaces(vertex):
        if volmesh.is_halfface_on_boundary(hfkey):
            vertex_hfkeys.append(hfkey)

    xyz = volmesh.vertex_coordinates(vertex)
    normal = volmesh.vertex_normal(vertex)

    def OnDynamicDraw(sender, e):
        cp = e.CurrentPoint
        for hfkey in vertex_hfkeys:
            for vertex in volmesh.halfface_vertices(hfkey):
                sp = volmesh.vertex_coordinates(vertex)
                e.Display.DrawLine(Point3d(*sp), cp, black, 2)

    ip = Point3d(*xyz)
    axis = Rhino.Geometry.Line(ip, ip + Vector3d(*normal))
    gp = get_target_point(axis, OnDynamicDraw)

    volmesh_vertex_lift(volmesh, vertex, gp, vertex_hfkeys)

    volmesh.draw()


def rhino_volmesh_vertex_merge(volmesh):
    """Rhino wrapper for the vertex merge operation.
    """

    volmesh.draw()

    keys = mesh_select_vertices(volmesh)
    xyz = centroid_points([volmesh.vertex_coordinates(key) for key in keys])

    volmesh_vertex_merge(volmesh, keys, xyz)

    volmesh.draw()


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   face operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def rhino_volmesh_halfface_pinch(volmesh):
    hfkeys = _select_boundary_halffaces(volmesh)

    key = hfkeys[0]

    center = volmesh.halfface_center(key)
    normal = volmesh.halfface_oriented_normal(key)
    line = (center, add_vectors(center, normal))

    # --------------------------------------------------------------------------
    #  dynamic draw
    # --------------------------------------------------------------------------
    rs.EnableRedraw(True)

    def OnDynamicDraw(sender, e):

        cp = e.CurrentPoint
        plane = (cp, normal)
        line = (center, add_vectors(center, normal))
        it = intersection_line_plane(line, plane)

        translation = subtract_vectors(it, center)
        dot = dot_vectors(normal, translation)
        dot = dot / abs(dot)

        dist = distance_point_point(center, it) * dot

        for hfkey in hfkeys:
            hf_center = volmesh.halfface_center(hfkey)
            hf_normal = volmesh.halfface_oriented_normal(hfkey)
            ep = add_vectors(hf_center, scale_vector(hf_normal, dist))
            e.Display.DrawDottedLine(Point3d(*hf_center), Point3d(*ep), feedback_color)
            e.Display.DrawPoint(Point3d(*ep), 0, 4, black)
            for vkey in volmesh.halfface_vertices(hfkey):
                sp = volmesh.vertex_coordinates(vkey)
                e.Display.DrawLine(Point3d(*sp), Point3d(*ep), black, 2)

    # --------------------------------------------------------------------------
    #  input point
    # --------------------------------------------------------------------------
    ip = Point3d(*center)
    axis = Rhino.Geometry.Line(ip, ip + Vector3d(*normal))
    gp = get_target_point(axis, OnDynamicDraw)

    plane = (gp, normal)
    it = intersection_line_plane(line, plane)

    translation = subtract_vectors(it, center)
    dot = dot_vectors(normal, translation)
    dot = dot / abs(dot)

    rise = distance_point_point(center, it) * dot

    for hfkey in hfkeys:
        hf_center = volmesh.halfface_center(hfkey)
        hf_normal = volmesh.halfface_oriented_normal(hfkey)
        ep = add_vectors(hf_center, scale_vector(hf_normal, rise))

        volmesh_halfface_pinch(volmesh, hfkey, ep)

    volmesh.draw()

    return volmesh


def rhino_volmesh_merge_adjacent_halffaces(volmesh):

    volmesh.draw()

    hfkeys = _select_boundary_halffaces(volmesh)
    volmesh_merge_adjacent_halffaces(volmesh, hfkeys)

    volmesh.draw()

    return volmesh


def rhino_volmesh_pull_halffaces(volmesh, hfkey, uniform=False):

    # # --------------------------------------------------------------------------
    # #  1. display boundary halffaces
    # # --------------------------------------------------------------------------
    # boundary_halffaces = volmesh.halffaces_on_boundaries()

    # volmesh.clear()
    # volmesh.draw_edges()
    # volmesh.draw_faces(faces=boundary_halffaces)

    # rs.EnableRedraw(True)

    # # --------------------------------------------------------------------------
    # #  2. select halfface and its dependents
    # # --------------------------------------------------------------------------
    # hf_inspector = VolmeshHalffaceInspector(volmesh,
    #                                         hfkeys=boundary_halffaces,
    #                                         dependents=True)
    # hf_inspector.enable()

    # hfkey = mesh_select_face(volmesh)

    # hf_inspector.disable()

    # del hf_inspector

    # hf dependent hfs

    dep_hfkeys = volmesh.halfface_manifold_neighborhood(hfkey, ring=50)

    # --------------------------------------------------------------------------
    #  3. move face
    # --------------------------------------------------------------------------
    ckey = volmesh.halfface_cell(hfkey)
    cell_vkeys = volmesh.cell_vertices(ckey)

    hf_vkeys = volmesh.halfface_vertices(hfkey)
    hf_normal = volmesh.halfface_normal(hfkey)
    hf_center = volmesh.halfface_center(hfkey)

    edges = {}
    for u in hf_vkeys:
        u_nbrs = volmesh.vertex_neighbors(u)
        for v in u_nbrs:
            if v in cell_vkeys:
                if v not in hf_vkeys:
                    edges[u] = v

    xyz_dict = {}
    for vkey in cell_vkeys:
        xyz_dict[vkey] = volmesh.vertex_coordinates(vkey)

    if uniform:
        hfkeys = dep_hfkeys + [hfkey]
        target_normal = _halffaces_avg_normals(volmesh, hfkeys)
    else:
        target_normal = None

    # dynamic draw -------------------------------------------------------------

    def OnDynamicDraw(sender, e):
        cp = e.CurrentPoint

        xyz = _volmesh_compute_dependent_face_intersections(volmesh, hfkey, cp, target_normal)

        seen = set()

        for fkey in dep_hfkeys + [hfkey]:
            hf_edges = volmesh.halfface_halfedges(fkey)
            for edge in hf_edges:
                u = edge[0]
                v = edge[1]
                pair = frozenset([u, v])
                if pair not in seen:
                    init_u = volmesh.vertex_coordinates(u)
                    init_u_xyz = Point3d(*init_u)
                    u_xyz = Point3d(*xyz[u])
                    v_xyz = Point3d(*xyz[v])
                    e.Display.DrawLine(Line(u_xyz, v_xyz), black, 4)
                    e.Display.DrawArrow(Line(init_u_xyz, u_xyz),
                                        arrow_color,
                                        12,
                                        0)
                    # e.Display.DrawDottedLine(init_u_xyz, u_xyz, feedback_color)

                    seen.add(pair)

        for u, v in volmesh.edges_iter():
            if frozenset([u, v]) not in seen:
                sp = volmesh.vertex_coordinates(u)
                ep = volmesh.vertex_coordinates(v)
                if u in xyz:
                    sp = xyz[u]
                if v in xyz:
                    ep = xyz[v]
                e.Display.DrawLine(Line(Point3d(*sp), Point3d(*ep)), black, 2)

    # --------------------------------------------------------------------------
    #  4. new face location
    # --------------------------------------------------------------------------
    ip = Point3d(*hf_center)
    line = Rhino.Geometry.Line(ip, ip + Vector3d(*hf_normal))
    gp = get_target_point(line, OnDynamicDraw)

    # --------------------------------------------------------------------------
    #  5. update halfface and its dependents
    # --------------------------------------------------------------------------
    new_xyz = _volmesh_compute_dependent_face_intersections(volmesh, hfkey, gp, target_normal)

    for key in new_xyz:
        coordinates = new_xyz[key]
        volmesh.vertex_update_xyz(key, coordinates, constrained=False)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cell operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def rhino_volmesh_cell_subdivide_barycentric(volmesh, formdiagram=None):

    cell_colors = {}
    ckeys = volmesh.cell.keys()
    for index, ckey in enumerate(ckeys):
        color = (0, 0, 0)
        if len(ckeys) != 1:
            value = float(index) / (len(ckeys) - 1)
            color = i_to_rgb(value)
        cell_colors[ckey] = color

    volmesh.clear()
    volmesh.draw_edges()
    volmesh.draw_faces()
    volmesh.draw_cell_labels(colors=True)
    rs.EnableRedraw(True)

    # dynamic selector
    cell_inspector = VolmeshCellInspector(volmesh, color_dict=cell_colors)
    cell_inspector.enable()
    sel_ckeys = CellSelector.select_cells(volmesh)
    cell_inspector.disable()
    del cell_inspector
    volmesh.clear_cell_labels()

    for ckey in sel_ckeys:
        volmesh_cell_subdivide_barycentric(volmesh, ckey)

    volmesh.clear()
    volmesh.draw()

    return volmesh


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   helpers
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def _volmesh_compute_dependent_face_intersections(volmesh,
                                                  hfkey,
                                                  xyz,

                                                  normal=None):

    if not normal:
        normal = None

    vertex_xyz = _cell_update_halfface(volmesh, hfkey, xyz, normal)

    dep_hfkeys = _boundary_halfface_edge_dependents(volmesh, hfkey)
    # dep_hfkeys = volmesh.halfface_manifold_neighbors(hfkey)
    hf_centers = {}

    for nbr_hfkey in dep_hfkeys:
        center_key = dep_hfkeys[nbr_hfkey]
        center_xyz = vertex_xyz[center_key]
        hf_centers[nbr_hfkey] = center_xyz

    dependents = set(_boundary_halfface_edge_dependents(volmesh, hfkey).keys())
    # dependents = set(volmesh.halfface_manifold_neighbors(hfkey))
    seen = set()

    i = 0
    while True:
        if i == 100:
            break
        if i != 0 and len(seen) == 0:
            break

        temp = []

        for d_hfkey in dependents:
            if d_hfkey not in seen:

                # compute new cell for the d_hfkey
                center = hf_centers[d_hfkey]
                next_xyz = _cell_update_halfface(volmesh,
                                                 d_hfkey,
                                                 center,
                                                 normal)
                for vkey in next_xyz:
                    if vkey not in vertex_xyz:
                        vertex_xyz[vkey] = next_xyz[vkey]

                next_d_hfkeys = _boundary_halfface_edge_dependents(volmesh, d_hfkey)

                for fkey in next_d_hfkeys:
                    if fkey not in hf_centers:
                        center_key = next_d_hfkeys[fkey]
                        center_xyz = vertex_xyz[center_key]
                        hf_centers[fkey] = center_xyz

                temp += next_d_hfkeys

                seen.add(d_hfkey)

        dependents.update(temp)
        i += 1

    if hfkey in dependents:
        dependents.remove(hfkey)

    return vertex_xyz


def _cell_update_halfface(volmesh, hfkey, xyz, normal=None):
    """Return new coordinates for the halfface.
    """

    new_cell_xyz = {}

    ckey = volmesh.halfface_cell(hfkey)
    cell_vkeys = volmesh.cell_vertices(ckey)
    hf_vkeys = volmesh.halfface_vertices(hfkey)
    if not normal:
        normal = volmesh.halfface_normal(hfkey)
    plane = (xyz, normal)

    edges = {key: [] for key in hf_vkeys}
    for u in hf_vkeys:
        nbrs = volmesh.vertex_neighbors(u)
        for v in nbrs:
            if v not in hf_vkeys and v in cell_vkeys:
                edges[u].append(v)

    for u in hf_vkeys:
        v = edges[u][0]
        u_xyz = volmesh.vertex_coordinates(u)
        v_xyz = volmesh.vertex_coordinates(v)
        line = (u_xyz, v_xyz)
        it = intersection_line_plane(line, plane)
        new_cell_xyz[u] = it

    return new_cell_xyz


def _volmesh_current_halfface_oriented_normals(volmesh):
    normals = {}
    for hfkey in volmesh.halfface:
        normal = volmesh.halfface_normal(hfkey)
        normals[hfkey] = normal
    return normals


def _volmesh_current_halfface_centers(volmesh):
    centers = {}
    for hfkey in volmesh.halfface:
        center = volmesh.halfface_center(hfkey)
        centers[hfkey] = center
    return centers


def _boundary_halfface_edge_dependents(volmesh, halfface):
    dep_hfkeys = {}
    # cell = volmesh.halfface_cell(halfface)
    # for halfedge in volmesh.halfface_halfedges(halfface):
    #     for face in volmesh.edge_halffaces(halfedge):
    #         if volmesh.is_halfface_on_boundary(face):
    #             if volmesh.halfface_cell(face) != cell:
    #                 dep_hfkeys[face] = halfedge[0]

    ckey = volmesh.halfface_cell(halfface)
    hf_edges = volmesh.halfface_halfedges(halfface)
    for edge in hf_edges:
        u = edge[0]
        v = edge[1]
        adj_hfkey = volmesh._cell[ckey][v][u]
        w = volmesh.halfface_vertex_ancestor(adj_hfkey, v)
        nbr_ckey = volmesh._plane[u][v][w]
        if nbr_ckey is not None:
            dep_hfkey = volmesh._cell[nbr_ckey][v][u]
            dep_hfkeys[dep_hfkey] = u
    return dep_hfkeys


def _boundary_halfface_manifold_neighbors(volmesh, halfface):
    '''intended for boundary faces only'''
    nbrs = []
    cell = volmesh.halfface_cell(halfface)
    for halfedge in volmesh.halfface_halfedges(halfface):
        for face in volmesh.edge_halffaces(halfedge):
            if volmesh.is_halfface_on_boundary(face):
                if face != halfface:
                    if volmesh.halfface_cell(face) != cell:
                        nbrs.append(face)
    # for u, v in volmesh.halfface_halfedges(halfface):
    #     nbr_halfface = volmesh._cell[cell][v][u]
    #     w = volmesh.halfface_vertex_ancestor(nbr_halfface, v)
    #     nbr_cell = volmesh._plane[u][v][w]
    #     if nbr_cell is not None:
    #         nbr = volmesh._cell[nbr_cell][v][u]
    #         nbrs.append(nbr)
    return nbrs


def _boundary_halfface_manifold_neighborhood(volmesh, hfkey, ring=1):
    """Return the halfface neighborhood of a halfface across their edges.
    Parameters
    ----------
    key : hashable
        The identifier of the halfface.
    Returns
    -------
    list
        The list of neighboring halffaces.
    Notes
    -----
    Neighboring halffaces on the same cell are not included.
    """
    nbrs = set(_boundary_halfface_manifold_neighbors(volmesh, hfkey))
    i = 1
    while True:
        if i == ring:
            break
        temp = []
        for nbr_hfkey in nbrs:
            temp += _boundary_halfface_manifold_neighbors(volmesh, hfkey)
        nbrs.update(temp)
        i += 1
    return list(nbrs - set([hfkey]))


def _halffaces_avg_normals(volmesh, hfkeys):
    vectors = []
    for hfkey in hfkeys:
        vectors.append(volmesh.halfface_normal(hfkey))
    return normalize_vector(centroid_points(vectors))


def _select_boundary_halffaces(volmesh):

    hfkeys = volmesh.halffaces_on_boundaries()

    volmesh.clear()
    volmesh.draw_edges()
    volmesh.draw_faces(fkeys=hfkeys)
    rs.EnableRedraw(True)

    hfkey = mesh_select_faces(volmesh)

    volmesh.draw()

    return hfkey


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas_rhino.artists import NetworkArtist


__all__ = ['NetworkArtist']


class NetworkArtist(NetworkArtist):
    """A customised `NetworkArtist` for 3GS `Network`-based data structures."""

    @property
    def diagram(self):
        """The diagram assigned to the artist."""
        return self.network

    @diagram.setter
    def diagram(self, diagram):
        self.network = diagram



"""
********************************************************************************
compas_3gs.operations
********************************************************************************

.. currentmodule:: compas_3gs.operations


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .celloperations import *  # noqa: F401 F403
from .volmeshoperations import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



"""
********************************************************************************
compas_3gs.rhino
********************************************************************************

.. currentmodule:: compas_3gs.rhino


Control
================================================================================

Functionalities to enhance user interactivity in Rhino.

Selectors
---------

Various functions for selecting compas_3gs objects in Rhino.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    CellSelector


Modifiers
---------


Inspectors
----------

Dynamic drawing and visualisation functions to enhance selectors and modifers.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    VolmeshVertexInspector
    VolmeshHalffaceInspector
    VolmeshCellInspector
    BiCellInspector


Helpers
-------

Helper functions for selectors, modifiers and inspectors.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    get_initial_point
    get_target_point

|

____

Display
================================================================================

Various drawing functions for Rhino.


Drawing
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    draw_egi_arcs

.. autosummary::
    :toctree: generated/
    :nosignatures:

    draw_cell_force_vectors
    draw_cell_labels
    clear_cell_labels

.. autosummary::
    :toctree: generated/
    :nosignatures:

    draw_network_external_forces
    draw_network_pipes
    draw_directed_hf_and_uv


.. autosummary::
    :toctree: generated/
    :nosignatures:

    draw_volmesh_face_normals

.. autosummary::
    :toctree: generated/
    :nosignatures:

    bake_cells_as_polysurfaces


Modes
-----


Conduits
--------


Helpers
-------


|

____

Wrappers
================================================================================

Wrapper functions specifically for use and visulisation in Rhino.


Algorithms
----------


Constructors
------------


Modifications
-------------


Operations
----------




"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .artists import *  # noqa: F401 F403
from .control import *  # noqa: F401 F403
from .display import *  # noqa: F401 F403
from .objects import *  # noqa: F401 F403
from .scene import *  # noqa: F401 F403
from .wrappers import *  # noqa: F401 F403
from .helpers import *  # noqa: F401 F403
from .forms import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



import os
import json
import compas
import importlib
import subprocess

PLUGIN_NAME = "3GS"


def run(cmd):
    # p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=compas._os.prepare_environment())
    stdout, stderr = p.communicate()
    return stdout, stderr, p.returncode


def get_plugin_path():
    plugin_path = os.path.dirname(__file__)
    plugin_path = os.path.realpath(plugin_path)
    plugin_path = os.path.join(plugin_path, "..", "..")
    plugin_path = os.path.abspath(plugin_path)

    return plugin_path


def get_register_json():
    register_json_path = os.path.join(compas.APPDATA, "compas_plugins.json")
    if os.path.exists(register_json_path):
        return json.load(open(register_json_path))
    else:
        return None


def check():

    if not compas.WINDOWS:
        print("checking skipped on non-windows systems")
        return True

    register_json = get_register_json()
    if not register_json:
        print("register json not found")
        return False

    print("Current plugin: ", register_json["Current"])
    if PLUGIN_NAME != register_json["Current"]:
        return False

    plugin_info = register_json["Plugins"][PLUGIN_NAME]

    print("Checking Dependencies")
    all_passed = True
    for name in plugin_info["packages"]:
        package = importlib.import_module(name)
        current = package.__version__
        required = plugin_info["packages"][name]["version"]
        required = required.split("-")[0]
        passed = current == required
        print(name, current, " -> ", required, "Passed:", passed)

        if not passed:
            all_passed = False
            break

    print("All passed:", all_passed)
    return all_passed


def activate():

    register_json = get_register_json()
    plugin_info = register_json["Plugins"][PLUGIN_NAME]

    out, err, code = run("%s -m compas_3gs.install" % plugin_info["python"])

    if code == 0:
        print(out.decode())
        print("Activation finished! Please restart rhino for changes to take effect!")
        return True
    else:
        print('exit with code', code)
        print(err.decode())
        return False


if __name__ == '__main__':

    if check():
        print("Current plugin is already activated")
    else:
        print("Env has changed, re-activating ", PLUGIN_NAME)
        activate()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

from copy import deepcopy

try:
    import rhinoscriptsyntax as rs
    import Rhino
except ImportError:
    compas.raise_if_ironpython()


__all__ = ['get_initial_point',
           'get_target_point',

           'set_target_areas']


def get_initial_point(message='Point to move from?'):
    ip = Rhino.Input.Custom.GetPoint()
    ip.SetCommandPrompt(message)
    ip.Get()
    ip = ip.Point()
    return ip


def get_target_point(constraint, OnDynamicDraw, option=None, message='Point to move to?'):
    gp = Rhino.Input.Custom.GetPoint()
    gp.SetCommandPrompt(message)

    gp.Constrain(constraint)

    if option is not None:
        gp.Constrain(constraint, option)

    gp.DynamicDraw += OnDynamicDraw
    gp.Get()
    gp = gp.Point()

    return gp


def set_target_areas(area_dict):

    go = Rhino.Input.Custom.GetOption()
    go.SetCommandPrompt("Enter target areas")

    # face key -----------------------------------------------------------------
    sortedkeys = sorted(area_dict.keys())
    key_strings = [str("key_" + str(key)) for key in sortedkeys]
    fkey_index = 0
    key_list = go.AddOptionList("pick_face", ["All"] + key_strings, fkey_index)

    # assign new target areas --------------------------------------------------
    new_area_dict = deepcopy(area_dict)

    while True:

        opt = go.Get()

        if go.CommandResult() != Rhino.Commands.Result.Success:
            break

        elif opt == Rhino.Input.GetResult.Option:  # keep picking options
            if go.OptionIndex() == key_list:
                fkey_index = int(go.Option().CurrentListOptionIndex)

                if fkey_index == 0:
                    avg = sum(area_dict.values()) / len(area_dict)

                    target_area = rs.GetReal("Enter target area for ALL faces", avg, 0, 1000.0)
                    if target_area:
                        for key in new_area_dict:
                            new_area_dict[key] = target_area

                else:
                    fkey_key = sortedkeys[fkey_index - 1]
                    current_area = area_dict[fkey_key]
                    target_area = rs.GetReal("Enter target area value", current_area, 0, 1000.0)
                    if target_area:
                        new_area_dict[fkey_key] = target_area

            # print current targets --------------------------------------------
            print('-----------------------------------------------------------')
            for fkey in new_area_dict:
                print("face", fkey, "has target area of :", new_area_dict[fkey])
            print('-----------------------------------------------------------')

            continue

        break

    return new_area_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



"""
********************************************************************************
compas_3gs
********************************************************************************

.. currentmodule:: compas_3gs

.. toctree::
    :maxdepth: 1

    compas_3gs.algorithms
    compas_3gs.datastructures
    compas_3gs.diagrams
    compas_3gs.operations
    compas_3gs.rhino
    compas_3gs.utilities

"""

from __future__ import print_function

import os


__author__ = ['Juney Lee', ]
__copyright__ = 'Copyright 2018 - Block Research Group, ETH Zurich'
__license__ = 'MIT License'
__email__ = 'juney.lee@arch.ethz.ch'

__version__ = '0.5.2'


HERE = os.path.dirname(__file__)
HOME = os.path.abspath(os.path.join(HERE, '../..'))
DATA = os.path.abspath(os.path.join(HERE, '../../data'))
TEMP = os.path.abspath(os.path.join(HERE, '../../temp'))

__all__ = []



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import compas

from compas.geometry import add_vectors

from compas_3gs.rhino.control import get_initial_point


try:
    import Rhino

    from Rhino.ApplicationSettings import ModelAidSettings
    from Rhino.Geometry import Point3d

    from System.Drawing.Color import FromArgb

    dotted_color = FromArgb(0, 0, 0)
    arrow_color = FromArgb(255, 0, 79)
    edge_color = FromArgb(0, 0, 0)

except ImportError:
    compas.raise_if_ironpython()


__all__ = ['rhino_vertex_modify_fixity',
           'rhino_vertex_move',
           'rhino_vertex_align']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh or network
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def rhino_vertex_modify_fixity(diagram, vertices):
    """Modifies the fixity attribute(s) of selected vertices.

    """

    go = Rhino.Input.Custom.GetOption()
    go.SetCommandPrompt('Set axes Constraints')

    boolOptionA = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
    boolOptionX = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
    boolOptionY = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
    boolOptionZ = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')

    go.AddOptionToggle('A', boolOptionA)
    go.AddOptionToggle('X', boolOptionX)
    go.AddOptionToggle('Y', boolOptionY)
    go.AddOptionToggle('Z', boolOptionZ)

    while True:
        opt = go.Get()
        if go.CommandResult() != Rhino.Commands.Result.Success:
            break
        if opt == Rhino.Input.GetResult.Option:  # keep picking options
            continue
        break

    if not vertices:
        return

    for vertex in vertices:
        if boolOptionA.CurrentValue:
            diagram.vertex_attribute(vertex, 'x_fix', True)
            diagram.vertex_attribute(vertex, 'y_fix', True)
            diagram.vertex_attribute(vertex, 'z_fix', True)

        else:
            diagram.vertex_attribute(vertex, 'x_fix', boolOptionX.CurrentValue)
            diagram.vertex_attribute(vertex, 'y_fix', boolOptionY.CurrentValue)
            diagram.vertex_attribute(vertex, 'z_fix', boolOptionZ.CurrentValue)


def rhino_vertex_move(diagram, vertices):
    """Moves the selected vertices.

    """

    nbr_vkeys = {}
    edges = set()
    for vertex in vertices:
        all_nbrs = diagram.vertex_neighbors(vertex)
        nbrs = []
        for nbr in all_nbrs:
            if nbr in vertices:
                edges.add(frozenset([vertex, nbr]))
            else:
                nbrs.append(nbr)
        nbr_vkeys[vertex] = nbrs

    ip = get_initial_point()

    def OnDynamicDraw(sender, e):
        cp = e.CurrentPoint
        translation = cp - ip
        for vertex in vertices:
            xyz = diagram.vertex_coordinates(vertex)
            sp = Point3d(*xyz)
            for nbr_vkey in nbr_vkeys[vertex]:
                nbr = diagram.vertex_coordinates(nbr_vkey)
                np = Point3d(*nbr)
                line = Rhino.Geometry.Line(sp, sp + translation)
                e.Display.DrawDottedLine(np, sp + translation, dotted_color)
                e.Display.DrawArrow(line, arrow_color, 15, 0)

        for pair in list(edges):
            pair = list(pair)
            u = diagram.vertex_coordinates(pair[0])
            v = diagram.vertex_coordinates(pair[1])
            sp = Point3d(*u) + translation
            ep = Point3d(*v) + translation
            e.Display.DrawLine(sp, ep, edge_color, 3)

    ModelAidSettings.Ortho = True
    gp = Rhino.Input.Custom.GetPoint()
    gp.DynamicDraw += OnDynamicDraw
    gp.SetCommandPrompt('Point to move to')
    ortho_option = Rhino.Input.Custom.OptionToggle(False, 'Off', 'On')
    gp.AddOptionToggle('ortho_snap', ortho_option)

    while True:
        ModelAidSettings.Ortho = ortho_option.CurrentValue
        get_rc = gp.Get()
        gp.SetBasePoint(ip, False)
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            continue
        if get_rc == Rhino.Input.GetResult.Option:
            continue
        elif get_rc == Rhino.Input.GetResult.Point:
            target = gp.Point()
        break

    translation = target - ip
    for vertex in vertices:
        new_xyz = add_vectors(diagram.vertex_coordinates(vertex), translation)
        diagram.vertex_update_xyz(vertex, new_xyz, constrained=False)


def rhino_vertex_align(diagram, vertices):

    def update_point(old, new):
        if boolOptionX.CurrentValue is True:
            old[0] = new[0]
        if boolOptionY.CurrentValue is True:
            old[1] = new[1]
        if boolOptionZ.CurrentValue is True:
            old[2] = new[2]
        return old

    # --------------------------------------------------------------------------
    # get vkeys to align
    # --------------------------------------------------------------------------

    nbr_vkeys = {}
    edges = set()
    for vertex in vertices:
        all_nbrs = diagram.vertex_neighbors(vertex)
        nbrs = []
        for nbr in all_nbrs:
            if nbr in vertices:
                edges.add(frozenset([vertex, nbr]))
            else:
                nbrs.append(nbr)
        nbr_vkeys[vertex] = nbrs

    # --------------------------------------------------------------------------
    # get rhino point
    # --------------------------------------------------------------------------
    gp = Rhino.Input.Custom.GetPoint()
    gp.SetCommandPrompt('Set alignment Constraints')

    boolOptionX = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
    boolOptionY = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
    boolOptionZ = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')

    gp.AddOptionToggle('X', boolOptionX)
    gp.AddOptionToggle('Y', boolOptionY)
    gp.AddOptionToggle('Z', boolOptionZ)

    def OnDynamicDraw(sender, e):
        cp = e.CurrentPoint

        for vertex in vertices:
            xyz = diagram.vertex_coordinates(vertex)
            sp = Point3d(*xyz)
            sp_f = update_point(sp, cp)
            for nbr_vkey in nbr_vkeys[vertex]:
                nbr = diagram.vertex_coordinates(nbr_vkey)
                np = Point3d(*nbr)
                e.Display.DrawDottedLine(np, sp_f, dotted_color)
                e.Display.DrawLine(sp, sp_f, edge_color, 3)

        for pair in list(edges):
            pair = list(pair)
            u = diagram.vertex_coordinates(pair[0])
            v = diagram.vertex_coordinates(pair[1])
            sp = update_point(Point3d(*u), cp)
            ep = update_point(Point3d(*v), cp)
            e.Display.DrawLine(sp, ep, edge_color, 3)

    while True:
        get_rc = gp.Get()
        gp.DynamicDraw += OnDynamicDraw
        # loop until a point is picked -----------------------------------------
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            break
        if get_rc == Rhino.Input.GetResult.Option:  # keep picking options
            continue
        # loop until a point is picked -----------------------------------------
        elif get_rc == Rhino.Input.GetResult.Point:
            target = gp.Point()
        break

    for vertex in vertices:
        xyz = update_point(diagram.vertex_coordinates(vertex), target)
        diagram.vertex_update_xyz(vertex, xyz, constrained=False)


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   network
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


# def network_vertex_fixity(network):

#     vkeys = mesh_select_vertices(network)

#     go = Rhino.Input.Custom.GetOption()
#     go.SetCommandPrompt('Set axes Constraints')

#     boolOptionA = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
#     boolOptionX = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
#     boolOptionY = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')
#     boolOptionZ = Rhino.Input.Custom.OptionToggle(False, 'False', 'True')

#     go.AddOptionToggle('A', boolOptionA)
#     go.AddOptionToggle('X', boolOptionX)
#     go.AddOptionToggle('Y', boolOptionY)
#     go.AddOptionToggle('Z', boolOptionZ)

#     while True:
#         opt = go.Get()
#         if go.CommandResult() != Rhino.Commands.Result.Success:
#             break
#         if opt == Rhino.Input.GetResult.Option:  # keep picking options
#             continue
#         break

#     if not vkeys:
#         return

#     for vkey in vkeys:
#         if boolOptionA.CurrentValue:
#             network.v_data[vkey]['x_fix'] = True
#             network.v_data[vkey]['y_fix'] = True
#             network.v_data[vkey]['z_fix'] = True
#         network.v_data[vkey]['x_fix'] = boolOptionX.CurrentValue
#         network.v_data[vkey]['y_fix'] = boolOptionY.CurrentValue
#         network.v_data[vkey]['z_fix'] = boolOptionZ.CurrentValue

#     network.draw()

#     return network


# def network_vertex_move(network):

#     vkeys = mesh_select_vertices(network)

#     nbr_vkeys = {}
#     edges = set()
#     for vkey in vkeys:
#         all_nbrs = network.vertex_neighbours(vkey)
#         nbrs = []
#         for nbr in all_nbrs:
#             if nbr in vkeys:
#                 edges.add(frozenset([vkey, nbr]))
#             else:
#                 nbrs.append(nbr)
#         nbr_vkeys[vkey] = nbrs

#     ip   = get_initial_point()

#     def OnDynamicDraw(sender, e):
#         cp = e.CurrentPoint
#         translation = cp - ip
#         for vkey in vkeys:
#             xyz = network.vertex_coordinates(vkey)
#             sp  = Point3d(*xyz)
#             for nbr_vkey in nbr_vkeys[vkey]:
#                 nbr  = network.vertex_coordinates(nbr_vkey)
#                 np   = Point3d(*nbr)
#                 line = Rhino.Geometry.Line(sp, sp + translation)
#                 e.Display.DrawDottedLine(np, sp + translation, dotted_color)
#                 e.Display.DrawArrow(line, arrow_color, 15, 0)

#         for pair in list(edges):
#             pair = list(pair)
#             u  = network.vertex_coordinates(pair[0])
#             v  = network.vertex_coordinates(pair[1])
#             sp = Point3d(*u) + translation
#             ep = Point3d(*v) + translation
#             e.Display.DrawLine(sp, ep, edge_color, 3)

#     ModelAidSettings.Ortho = True
#     gp = Rhino.Input.Custom.GetPoint()
#     gp.DynamicDraw += OnDynamicDraw
#     gp.SetCommandPrompt('Point to move to')
#     ortho_option = Rhino.Input.Custom.OptionToggle(True, 'Off', 'On')
#     gp.AddOptionToggle('ortho_snap', ortho_option)

#     while True:
#         ModelAidSettings.Ortho = ortho_option.CurrentValue
#         get_rc = gp.Get()
#         gp.SetBasePoint(ip, False)
#         if gp.CommandResult() != Rhino.Commands.Result.Success:
#             continue
#         if get_rc == Rhino.Input.GetResult.Option:
#             continue
#         elif get_rc == Rhino.Input.GetResult.Point:
#             target = gp.Point()
#         break

#     translation = target - ip
#     for vkey in vkeys:
#         new_xyz = add_vectors(network.vertex_coordinates(vkey), translation)
#         network.vertex_update_xyz(vkey, new_xyz, constrained=False)

#     network.draw()


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



"""
********************************************************************************
compas_3gs.diagrams
********************************************************************************

.. currentmodule:: compas_3gs.diagrams


Polyhedral Cell
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    EGI
    Cell

|

____


Multi-cell Polyhedron
=====================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ForceVolMesh
    FormNetwork
    FormVolMesh

|

____


Cell Network
============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    CellNetwork


"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .egi import *  # noqa: F401 F403
from .cell import *  # noqa: F401 F403
from .polyhedral import *  # noqa: F401 F403
from .disjointed import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino


__commandname__ = "TGS__scene_clear"


def RunCommand(is_interactive):

    scene = sc.sticky['3GS']['scene']
    if not scene:
        return

    options = ["Yes", "No"]
    option = compas_rhino.rs.GetString("Clear all 3GS objects?", strings=options, defaultString="No")
    if not option:
        return

    if option == "Yes":
        scene.clear_layers()
        scene.purge()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    RunCommand(True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .attributes import *  # noqa: F401 F403
from .scene import *  # noqa: F401 F403
from .settings import *  # noqa: F401 F403

__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import ast
import compas

try:
    import rhinoscriptsyntax as rs
except ImportError:
    compas.raise_if_ironpython()


__all__ = ['VolMeshSelector',
           'CellSelector']


class VolMeshSelector(object):

    @staticmethod
    def select_vertex(self, message="Select a vertex."):
        guid = rs.GetObject(message, preselect=True, filter=rs.filter.point | rs.filter.textdot)
        if guid:
            prefix = self.attributes['name']
            name = rs.ObjectName(guid).split('.')
            if 'vertex' in name:
                if not prefix or prefix in name:
                    key = name[-1]
                    return ast.literal_eval(key)
        return None

    @staticmethod
    def select_vertices(self, message="Select vertices."):
        keys = []
        guids = rs.GetObjects(message, preselect=True, filter=rs.filter.point | rs.filter.textdot)
        if guids:
            prefix = self.attributes['name']
            seen = set()
            for guid in guids:
                name = rs.ObjectName(guid).split('.')
                if 'vertex' in name:
                    if not prefix or prefix in name:
                        key = name[-1]
                        if not seen.add(key):
                            key = ast.literal_eval(key)
                            keys.append(key)
        return keys

    @staticmethod
    def select_halfface(self, message="Select a halfface."):
        guid = rs.GetObject(message, preselect=True, filter=rs.filter.mesh)
        if guid:
            prefix = self.attributes['name']
            name = rs.ObjectName(guid).split('.')
            if 'face' in name:
                if not prefix or prefix in name:
                    key = name[-1]
                    return ast.literal_eval(key)
        return None

    @staticmethod
    def select_halffaces(self, message="Select halffaces."):
        keys = []
        guids = rs.GetObjects(message, preselect=True, filter=rs.filter.mesh)
        if guids:
            prefix = self.attributes['name']
            seen = set()
            for guid in guids:
                name = rs.ObjectName(guid).split('.')
                if 'face' in name:
                    if not prefix or prefix in name:
                        key = name[-1]
                        if not seen.add(key):
                            key = ast.literal_eval(key)
                            keys.append(key)
        return keys


class CellSelector(object):

    @staticmethod
    def select_cell(self, message="Select a cell."):
        rs.EnableRedraw(True)
        guid = rs.GetObject(message, preselect=True, filter=rs.filter.point | rs.filter.textdot)
        if guid:
            prefix = self.attributes['name']
            name = rs.ObjectName(guid).split('.')
            if 'cell' in name:
                if not prefix or prefix in name:
                    key = name[-1]
                    return ast.literal_eval(key)
        return None

    @staticmethod
    def select_cells(self, message="Select cells."):
        rs.EnableRedraw(True)
        keys = []
        guids = rs.GetObjects(message, preselect=True, filter=rs.filter.point | rs.filter.textdot)
        if guids:
            prefix = self.attributes['name']
            seen = set()
            for guid in guids:
                name = rs.ObjectName(guid).split('.')
                if 'cell' in name:
                    if not prefix or prefix in name:
                        key = name[-1]
                        if not seen.add(key):
                            key = ast.literal_eval(key)
                            keys.append(key)
        return keys


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



"""
********************************************************************************
compas_3gs.algorithms
********************************************************************************

.. currentmodule:: compas_3gs.algorithms


EGI
===

.. autosummary::
    :toctree: generated/
    :nosignatures:

    egi_from_vectors


____


Duality
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    volmesh_dual_volmesh
    volmesh_dual_network


____


Reciprocation
=============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    volmesh_reciprocate


____


Planarisation
=============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    volmesh_planarise


Arearisation
============


.. autosummary::
    :toctree: generated/
    :nosignatures:

    cell_arearise_face


____


Unified Diagram
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    volmesh_ud
    cellnetwork_ud

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .arearisation import *  # noqa: F401 F403
from .duality import *  # noqa: F401 F403
from .egi import *  # noqa: F401 F403
from .other import *  # noqa: F401 F403
from .planarisation import *  # noqa: F401 F403
from .reciprocation import *  # noqa: F401 F403
from .unifieddiagram import *  # noqa: F401 F403


__all__ = [name for name in dir() if not name.startswith('_')]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

import scriptcontext as sc

import compas_rhino

from compas_3gs.algorithms import volmesh_ud

from compas_3gs.utilities import get_force_colors_uv

try:
    import rhinoscriptsyntax as rs
except ImportError:
    compas.raise_if_ironpython()


__commandname__ = "TGS_unified_diagram"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('force')
    if not objects:
        compas_rhino.display_message("There is no force diagram in the scene.")
        return
    force = objects[0]

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('form')
    if not objects:
        compas_rhino.display_message("There is no form diagram in the scene.")
        return
    form = objects[0]

    # check global constraints -------------------------------------------------

    force.check_eq()
    form.check_eq()

    if not force.settings['_is.valid'] or not form.settings['_is.valid']:
        options = ["Yes", "No"]
        option = compas_rhino.rs.GetString("System is not in equilibrium... proceed?", strings=options, defaultString="No")
        if not option:
            return
        if option == "No":
            return

    show_loads = form.settings['show.loads']
    form.settings['show.loads'] = False

    # unified diagram ----------------------------------------------------------
    while True:

        rs.EnableRedraw(True)

        alpha = rs.GetReal('unified diagram scale', minimum=0.01, maximum=1.0)

        if alpha is None:
            break

        if not alpha:
            break

        compas_rhino.clear_layer(force.layer)
        compas_rhino.clear_layer(form.layer)

        # 1. get colors --------------------------------------------------------
        hf_color = (0, 0, 0)

        uv_c_dict = get_force_colors_uv(force.diagram, form.diagram, gradient=True)

        # 2. compute unified diagram geometries --------------------------------
        cells, prisms = volmesh_ud(force.diagram, form.diagram, scale=alpha)

        # 3. draw --------------------------------------------------------------
        for cell in cells:
            vertices = cells[cell]['vertices']
            faces = cells[cell]['faces']
            compas_rhino.draw_mesh(vertices, faces, layer=force.layer, name=str(cell), color=hf_color, redraw=False)

        for edge in prisms:
            vertices = prisms[edge]['vertices']
            faces = prisms[edge]['faces']
            compas_rhino.draw_mesh(vertices, faces, layer=force.layer, name=str(edge), color=uv_c_dict[edge], redraw=False)

        form.artist.draw_edges(color=uv_c_dict)

    form.settings['show.loads'] = show_loads

    scene.save()

# ==============================================================================
# Main
# ==============================================================================


if __name__ == '__main__':

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas

import scriptcontext as sc

import compas_rhino

from compas_3gs.diagrams.polyhedral import ForceVolMesh

from compas_rhino.utilities import volmesh_from_polysurfaces

try:
    import rhinoscriptsyntax as rs
except ImportError:
    compas.raise_if_ironpython()


__commandname__ = "TGS_force_from_polysurfaces"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('force')
    if objects:
        compas_rhino.display_message("Force diagram already exists in the scene!")
        return

    guids = rs.GetObjects("select polysurfaces", filter=rs.filter.polysurface)
    if not guids:
        return

    compas_rhino.rs.HideObjects(guids)

    force = volmesh_from_polysurfaces(ForceVolMesh, guids)

    scene.purge()
    scene.add_forcevolmesh(force, name='force', layer='3GS::ForceDiagram')

    objects = scene.find_by_name('force')
    force = objects[0]

    force.check_eq()

    scene.update()
    scene.save()

    print('Polyhedral force diagram successfully created.')


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino

from compas_rhino.objects import NetworkObject


__all__ = ['NetworkObject']


class NetworkObject(NetworkObject):
    """A customised `NetworkArtist` for 3GS `Network`-based data structures."""

    @property
    def diagram(self):
        """The diagram associated with the object."""
        return self.network

    @diagram.setter
    def diagram(self, diagram):
        self.network = diagram

    def select_node(self, message="Select vertices."):
        """Select vertices of the volmesh.

        Returns
        -------
        list
            A list of vertex identifiers.
        """
        guid = compas_rhino.select_point(message=message)
        if guid and guid in self.guid_node:
            return self.guid_node[guid]

    select_vertex = select_node
    select_vertices = NetworkObject.select_nodes



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import OrderedDict
import scriptcontext as sc

import Eto.Drawing as drawing
import Eto.Forms as forms
import Rhino.UI


__all__ = ["SettingsForm", "Settings_Tab"]


def get_scene():
    return sc.sticky['3GS']['scene']


class Settings_Tab(forms.TabPage):

    @classmethod
    def from_settings(cls, object_name, settings):
        tab = cls()
        tab.Text = object_name
        layout = forms.StackLayout()
        layout.Spacing = 2
        layout.HorizontalContentAlignment = forms.HorizontalAlignment.Stretch
        tab.Content = layout

        # link to original setting and keep a temporary new settings holder
        tab.settings = settings
        tab.new_settings = settings.copy()

        # create sections
        sections = {}
        for key in settings:
            split = key.split('.')
            prefix = split[0]
            postfix = '.'.join(split[1:])

            if prefix[0] == "_":
                continue

            if prefix not in sections:
                sections[prefix] = {}
            sections[prefix][postfix] = (key, settings[key])

        for prefix in sections:

            groupbox = forms.GroupBox(Text=prefix)
            groupbox.Padding = drawing.Padding(5)
            grouplayout = forms.DynamicLayout()
            grouplayout.Spacing = drawing.Size(3, 3)

            postfixies = list(sections[prefix].keys())
            postfixies.sort()
            for postfix in postfixies:
                key, value = sections[prefix][postfix]

                if type(value) == bool:
                    control = forms.CheckBox()
                    control.Checked = value
                    control.CheckedChanged += tab.EditEvent(key)
                elif (type(value) == list or type(value) == tuple) and len(value) == 3:
                    control = forms.ColorPicker()
                    control.Value = drawing.Color.FromArgb(*value)
                    control.ValueChanged += tab.EditEvent(key)
                elif type(value) == float or type(value) == int:
                    control = forms.NumericUpDown()
                    if type(value) == float:
                        digits = len(str(value).split('.')[-1])
                        control.DecimalPlaces = (digits + 2)
                        control.Increment = 0.1 ** (digits + 2)
                    control.Value = value
                    control.ValueChanged += tab.EditEvent(key)
                else:
                    control = forms.TextBox()
                    control.Text = str(value)
                    control.TextChanged += tab.EditEvent(key)

                label = forms.Label(Text=postfix)
                if postfix != '':
                    grouplayout.AddRow(label, None, control)
                else:
                    grouplayout.AddRow(control)

            groupbox.Content = grouplayout
            layout.Items.Add(groupbox)

        return tab

    def EditEvent(self, key):
        def on_edited(sender, event):
            try:
                if isinstance(sender, forms.ColorPicker):
                    color = sender.Value
                    new_value = [int(color.Rb), int(color.Gb), int(color.Bb)]
                elif isinstance(sender, forms.CheckBox):
                    new_value = sender.Checked
                elif isinstance(sender, forms.NumericUpDown):
                    new_value = sender.Value
                elif isinstance(sender, forms.TextBox):
                    new_value = sender.Text

                self.new_settings.update({key: new_value})

            except Exception as e:
                print(e)
        return on_edited

    def apply(self):
        self.settings.update(self.new_settings)


class SettingsForm(forms.Dialog[bool]):

    @classmethod
    def from_scene(cls, scene, object_types=None, global_settings=None):

        all_settings = OrderedDict()

        if global_settings:
            for name in global_settings:
                all_settings[name] = scene.settings[name]

        # pre-populate class default settings
        if object_types:
            for object_type in object_types:  # avoid property objects
                all_settings[object_type.__name__] = object_type.SETTINGS

        # overwite with object setting if added as node
        for key in scene.objects:
            node = scene.objects[key]
            node_type = node.__class__.__name__
            if node_type in all_settings:
                all_settings[node_type] = node.settings

        settingsForm = cls()
        settingsForm.scene = scene
        settingsForm.setup(all_settings)
        Rhino.UI.EtoExtensions.ShowSemiModal(settingsForm, Rhino.RhinoDoc.ActiveDoc, Rhino.UI.RhinoEtoApp.MainWindow)
        return settingsForm

    @classmethod
    def from_settings(cls, settings, name):

        all_settings = {name: settings}
        settingsForm = cls()
        settingsForm.setup(all_settings)
        Rhino.UI.EtoExtensions.ShowSemiModal(settingsForm, Rhino.RhinoDoc.ActiveDoc, Rhino.UI.RhinoEtoApp.MainWindow)
        return settingsForm

    def setup(self, all_settings):

        self.Title = "Settings"
        self.TabControl = self.tabs_from_settings(all_settings)
        tab_items = forms.StackLayoutItem(self.TabControl, True)
        layout = forms.StackLayout()
        layout.Spacing = 5
        layout.HorizontalContentAlignment = forms.HorizontalAlignment.Stretch
        layout.Items.Add(tab_items)

        sub_layout = forms.DynamicLayout()
        sub_layout.Spacing = drawing.Size(5, 0)
        sub_layout.AddRow(None, self.ok, self.cancel, self.apply)
        layout.Items.Add(forms.StackLayoutItem(sub_layout))

        self.Content = layout
        self.Padding = drawing.Padding(12)
        self.Resizable = True

    def tabs_from_settings(self, all_settings):
        control = forms.TabControl()
        control.TabPosition = forms.DockPosition.Top

        for object_name in all_settings:
            tab = Settings_Tab.from_settings(object_name, all_settings[object_name])
            control.Pages.Add(tab)

        return control

    @property
    def ok(self):
        self.DefaultButton = forms.Button(Text='OK')
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        self.AbortButton = forms.Button(Text='Cancel')
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    @property
    def apply(self):
        self.ApplyButton = forms.Button(Text='Apply')
        self.ApplyButton.Click += self.on_apply
        return self.ApplyButton

    def on_ok(self, sender, event):
        try:
            for page in self.TabControl.Pages:
                page.apply()
            if hasattr(self, 'scene'):
                self.scene.update()
        except Exception as e:
            print(e)
        self.Close()

    def on_apply(self, sender, event):
        try:
            for page in self.TabControl.Pages:
                page.apply()
            if hasattr(self, 'scene'):
                self.scene.update()
        except Exception as e:
            print(e)

    def on_cancel(self, sender, event):
        self.Close()


if __name__ == "__main__":

    from compas_ags.rhino import FormObject
    from compas_ags.rhino import ForceObject
    scene = get_scene()
    SettingsForm.from_scene(scene, object_types=[FormObject, ForceObject])
    # SettingsForm.from_settings(scene.settings, "solver")



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino

from compas_3gs.rhino import rhino_vertex_move


__commandname__ = "TGS_form_move_vertices"


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('form')
    if not objects:
        compas_rhino.display_message("There is no form diagram in the scene.")
        return
    form = objects[0]

    # --------------------------------------------------------------------------

    current_setting = form.settings['show.nodes']
    if not current_setting:
        form.settings['show.nodes'] = True
        scene.update()

    vertices = form.select_vertices()

    rhino_vertex_move(form.diagram, vertices)

    form.settings['show.nodes'] = current_setting

    # --------------------------------------------------------------------------

    form.diagram.update_angle_deviations()

    objects = scene.find_by_name('force')
    if not objects:
        form.check_eq()
        scene.update()
        return
    force = objects[0]

    force.check_eq()
    form.check_eq()

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import ast

import rhinoscriptsyntax as rs
import scriptcontext as sc
import Eto.Drawing as drawing
import Eto.Forms as forms
import Rhino

find_object = sc.doc.Objects.Find


__all__ = ["AttributesForm"]


def get_scene():
    return sc.sticky['3GS']['scene']


class Tree_Table(forms.TreeGridView):
    def __init__(self, ShowHeader=True, sceneNode=None, table_type=None):
        self.ShowHeader = ShowHeader
        self.Height = 300
        self.last_sorted_to = None
        self.table_type = table_type
        self.sceneNode = sceneNode
        self.to_update = {}
        # self.AllowMultipleSelection=True

        # give colors to each item cells according to object settings
        color = {}
        if sceneNode and table_type:
            settings = sceneNode.settings

            # print(sceneNode.settings)

            # update general color settings for this table
            general_setting_key = "color.%s" % table_type

            if getattr(sceneNode.diagram, table_type):
                color.update({str(key): settings.get(general_setting_key) for key in getattr(sceneNode.diagram, table_type)()})

            # gather and update subsettings
            for full_setting_key in settings:
                sub_setting_str = full_setting_key.split(":")
                if len(sub_setting_str) > 1 and sub_setting_str[0] == general_setting_key:
                    sub_setting_key = sub_setting_str[-1]
                    items_where = getattr(sceneNode.diagram, '%s_where' % table_type)
                    color.update({str(key): settings.get(full_setting_key) for key in items_where({sub_setting_key: True})})
                    color.update({str(key): settings.get(full_setting_key) for key in items_where({'_' + sub_setting_key: True})})  # including read-only ones
                    color.update({str(key): settings.get("color.edges:is_ind") for key in items_where({"is_ind": True})})  # overwrite is_ind lastly

            # Additional settings for AGS
            tol = 0.001
            for edge in sceneNode.diagram.edges_where({'is_external': False}):
                if sceneNode.diagram.edge_attribute(edge, 'f') > + tol:
                    color[str(edge)] = settings['color.tension']
                if sceneNode.diagram.edge_attribute(edge, 'f') < - tol:
                    color[str(edge)] = settings['color.compression']

        def OnCellFormatting(sender, e):
            try:
                attr = e.Column.HeaderText

                # if not e.Column.Editable and attr != 'key':
                #     e.ForegroundColor = drawing.Colors.DarkGray

                if attr == 'EdgeLabel':
                    key = e.Item.Values[1]
                    if key in color:
                        rgb = color[key]
                        rgb = [c / 255. for c in rgb]
                        e.BackgroundColor = drawing.Color(*rgb)

            except Exception as exc:
                print('formating error', exc)

        self.CellFormatting += OnCellFormatting

    @classmethod
    def create_force_table(cls, sceneNode, dual=[]):
        datastructure = sceneNode.diagram
        table = cls(sceneNode=sceneNode, table_type='edges')
        table.add_column('EdgeLabel')
        table.add_column('Vertices')

        attributes = list(datastructure.default_edge_attributes.keys())
        attributes = table.sort_attributes(attributes)

        Allowed = ["is_ind", "f", "is_external"]
        attributes = filter(lambda attr: attr in Allowed, attributes)

        for attr in attributes:
            checkbox = type(datastructure.default_edge_attributes[attr]) == bool
            attr = attr.replace("_", "-")
            table.add_column(attr, Editable=False, checkbox=checkbox)

        treecollection = forms.TreeGridItemCollection()
        for index, edge in enumerate(datastructure.edges()):
            values = [index, str(edge)]

            for attr in attributes:
                value = datastructure.edge_attribute(edge, attr)
                if isinstance(value, float):
                    value = float("%.4g" % (value))
                values.append(value)

            edge_item = forms.TreeGridItem(Values=tuple(values))
            treecollection.Add(edge_item)
            # for key in edge:
            #     vertex_item = forms.TreeGridItem(Values=(str(key),))
            #     edge_item.Children.Add(vertex_item)

        table.DataStore = treecollection
        table.Activated += table.SelectEvent(sceneNode, 'guid_edge', dual=dual)
        table.ColumnHeaderClick += table.HeaderClickEvent()
        table.CellEdited += table.EditEvent()
        return table

    @classmethod
    def create_vertices_table(cls, sceneNode):
        datastructure = sceneNode.datastructure
        table = cls(sceneNode=sceneNode, table_type='vertices')
        table.add_column('key')
        attributes = list(datastructure.default_vertex_attributes.keys())
        attributes = table.sort_attributes(attributes)
        for attr in attributes:
            editable = attr[0] != '_'
            checkbox = type(datastructure.default_vertex_attributes[attr]) == bool
            if not editable:
                attr = attr[1:]
            table.add_column(attr, Editable=editable, checkbox=checkbox)
        treecollection = forms.TreeGridItemCollection()
        for key in datastructure.vertices():
            values = [str(key)]
            for attr in attributes:
                values.append(str(datastructure.vertex_attribute(key, attr)))
            treecollection.Add(forms.TreeGridItem(Values=tuple(values)))
        table.DataStore = treecollection
        table.Activated += table.SelectEvent(sceneNode, 'guid_vertex')
        table.ColumnHeaderClick += table.HeaderClickEvent()
        table.CellEdited += table.EditEvent()
        return table

    @classmethod
    def create_edges_table(cls, sceneNode):
        datastructure = sceneNode.datastructure
        table = cls(sceneNode=sceneNode, table_type='edges')
        table.add_column('key')
        attributes = list(datastructure.default_edge_attributes.keys())
        attributes = table.sort_attributes(attributes)

        for attr in attributes:
            editable = attr[0] != '_'
            checkbox = type(datastructure.default_edge_attributes[attr]) == bool
            if not editable:
                attr = attr[1:]
            table.add_column(attr, Editable=editable, checkbox=checkbox)

        treecollection = forms.TreeGridItemCollection()
        for key, edge in enumerate(datastructure.edges()):
            values = [str(edge)]
            for attr in attributes:
                values.append(datastructure.edge_attribute(edge, attr))
            edge_item = forms.TreeGridItem(Values=tuple(values))
            treecollection.Add(edge_item)
            for key in edge:
                vertex_item = forms.TreeGridItem(Values=(str(key),))
                edge_item.Children.Add(vertex_item)
        table.DataStore = treecollection
        table.Activated += table.SelectEvent(sceneNode, 'guid_edge', 'guid_vertex')
        table.ColumnHeaderClick += table.HeaderClickEvent()
        table.CellEdited += table.EditEvent()
        return table

    def sort_attributes(self, attributes):
        sorted_attributes = attributes[:]
        sorted_attributes.sort()

        switch = len(sorted_attributes)
        for i, attr in enumerate(sorted_attributes):
            if attr[0] != '_':
                switch = i
                break
        sorted_attributes = sorted_attributes[switch:] + sorted_attributes[:switch]

        return sorted_attributes

    def SelectEvent(self, sceneNode, guid_field, children_guid_field=None, dual=None):
        def on_selected(sender, event):
            try:
                rs.UnselectAllObjects()
                key = event.Item.Values[1]
                guid2key = getattr(sceneNode, guid_field)
                key2guid = {str(guid2key[guid]): guid for guid in guid2key}
                if key in key2guid:
                    find_object(key2guid[key]).Select(True)

                if dual:
                    force_edge_index = int(event.Item.Values[0])
                    force_edge_key2index = dual.diagram.edge_index(dual.diagram.dual)
                    force_edge_index2key = {force_edge_key2index[key]: key for key in force_edge_key2index}
                    force_edge_key = force_edge_index2key[force_edge_index]

                    guid2key = getattr(dual, guid_field)
                    key2guid = {guid2key[guid]: guid for guid in guid2key}
                    key2guid.update({(v, u): key2guid[(u, v)] for u, v in key2guid})
                    if force_edge_key in key2guid:
                        find_object(key2guid[force_edge_key]).Select(True)

                # elif children_guid_field:
                #     if key == '':
                #         key = event.Item.Values[1]
                #     guid2key = getattr(sceneNode.artist, children_guid_field)
                #     key2guid = {str(guid2key[guid]): guid for guid in guid2key}
                #     find_object(key2guid[key]).Select(True)

                # print('selected', key, key2guid[key])
                rs.Redraw()
            except Exception as e:
                print(e)

        return on_selected

    def EditEvent(self):
        def on_edited(sender, event):
            try:
                key = event.Item.Values[0]
                value = event.Item.Values[event.Column]
                attr = self.Columns[event.Column].HeaderText

                if self.table_type == 'vertices':
                    get_set_attributes = getattr(self.sceneNode.datastructure, 'vertex_attribute')
                if self.table_type == 'edges':
                    get_set_attributes = getattr(self.sceneNode.datastructure, 'edge_attribute')
                if self.table_type == 'faces':
                    get_set_attributes = getattr(self.sceneNode.datastructure, 'face_attribute')

                try:
                    new_value = ast.literal_eval(str(value))  # checkboxes value type is bool, turn them into str first to be parsed back to bool
                except Exception:
                    new_value = str(value)

                # convert key from str to original type: int,tuple...
                try:
                    key = ast.literal_eval(key)
                except Exception:
                    pass

                original_value = get_set_attributes(key, attr)

                if type(original_value) == float and type(new_value) == int:
                    new_value = float(new_value)
                if new_value != original_value:
                    if type(new_value) == type(original_value):
                        print('will update key: %s, attr: %s, value: %s' % (key, attr, new_value))
                        self.to_update[(key, attr)] = (get_set_attributes, new_value)
                    else:
                        print('invalid value type, needs: %s, got %s instead' % (type(original_value), type(new_value)))
                        event.Item.Values[event.Column] = original_value
                else:
                    print('value not changed from', original_value)

            except Exception as e:
                print("cell edit failed:", type(e), e)
        return on_edited

    def HeaderClickEvent(self):
        def on_hearderClick(sender, event):
            try:
                # print(event.Column.HeaderText)
                sender.sort(event.Column.HeaderText)
            except Exception as e:
                print(e)
        return on_hearderClick

    def add_column(self, HeaderText=None, Editable=False, checkbox=False):
        column = forms.GridColumn()
        if self.ShowHeader:
            column.HeaderText = HeaderText
        column.Editable = Editable
        if not checkbox:
            column.DataCell = forms.TextBoxCell(self.Columns.Count)
        else:
            column.DataCell = forms.CheckBoxCell(self.Columns.Count)
        column.Sortable = True
        self.Columns.Add(column)

    def sort(self, key):
        headerTexts = [column.HeaderText for column in self.Columns]
        index = headerTexts.index(key)
        print(key, index)

        def getValue(item):
            try:
                value = ast.literal_eval(item.Values[index])
            except (ValueError, TypeError):
                value = item.Values[index]
            return value

        items = sorted(self.DataStore, key=getValue, reverse=True)
        if self.last_sorted_to == key:
            items.reverse()
            self.last_sorted_to = None
        else:
            self.last_sorted_to = key

        self.DataStore = forms.TreeGridItemCollection(items)

    def apply(self):
        for _key in self.to_update:
            get_set_attributes, new_value = self.to_update[_key]
            key, attr = _key
            get_set_attributes(key, attr, new_value)


class Tree_Tab(forms.TabPage):

    @classmethod
    def from_sceneNode(cls, sceneNode, tab_type, dual=None):
        tab = cls()
        tab.Text = tab_type
        # create_table = getattr(Tree_Table, "create_%s_table" % tab_type)
        tab.Content = Tree_Table.create_force_table(sceneNode, dual=dual)
        return tab

    def apply(self):
        self.Content.apply()


class AttributesForm(forms.Dialog[bool]):

    @classmethod
    def from_sceneNode(cls, sceneNode, dual=None):
        attributesForm = cls()
        attributesForm.setup(sceneNode, dual=dual)
        Rhino.UI.EtoExtensions.ShowSemiModal(attributesForm, Rhino.RhinoDoc.ActiveDoc, Rhino.UI.RhinoEtoApp.MainWindow)
        return attributesForm

    def setup(self, sceneNode, dual=None):
        self.Title = "Attributes"
        self.sceneNode = sceneNode

        control = forms.TabControl()
        control.TabPosition = forms.DockPosition.Top

        tab = Tree_Tab.from_sceneNode(sceneNode, 'Edges', dual=dual)
        control.Pages.Add(tab)

        self.TabControl = control

        tab_items = forms.StackLayoutItem(self.TabControl, True)
        layout = forms.StackLayout()
        layout.Spacing = 5
        layout.HorizontalContentAlignment = forms.HorizontalAlignment.Stretch
        layout.Items.Add(tab_items)

        sub_layout = forms.DynamicLayout()
        sub_layout.Spacing = drawing.Size(5, 0)
        # sub_layout.AddRow(None, self.ok, self.cancel, self.apply)
        sub_layout.AddRow(None, self.cancel)
        layout.Items.Add(forms.StackLayoutItem(sub_layout))

        self.Content = layout
        self.Padding = drawing.Padding(12)
        self.Resizable = True
        self.ClientSize = drawing.Size(400, 600)

    @property
    def ok(self):
        self.DefaultButton = forms.Button(Text='OK')
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        # self.AbortButton = forms.Button(Text='Cancel')
        self.AbortButton = forms.Button(Text='Close')
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    @property
    def apply(self):
        self.ApplyButton = forms.Button(Text='Apply')
        self.ApplyButton.Click += self.on_apply
        return self.ApplyButton

    def on_ok(self, sender, event):
        try:
            for page in self.TabControl.Pages:
                if hasattr(page, 'apply'):
                    page.apply()
            get_scene().update()
        except Exception as e:
            print(e)
        self.Close()

    def on_apply(self, sender, event):
        try:
            for page in self.TabControl.Pages:
                if hasattr(page, 'apply'):
                    page.apply()
            get_scene().update()
        except Exception as e:
            print(e)

    def on_cancel(self, sender, event):
        self.Close()


if __name__ == "__main__":

    scene = get_scene()
    form = scene.find_by_name("Form")[0]
    force = scene.find_by_name("Force")[0]
    AttributesForm.from_sceneNode(form, dual=force)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from compas.datastructures import mesh_dual

from compas.geometry import dot_vectors
from compas.geometry import distance_point_point
from compas.geometry import midpoint_point_point
from compas.geometry import intersection_line_plane

from compas_3gs.diagrams import EGI


__all__ = ['cell_split_indet_face_vertices',
           'cell_collapse_short_edge',

           'cell_relocate_face',
           'cell_face_subdivide_barycentric',
           'cell_merge_coplanar_adjacent_faces']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cell vertices
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def cell_split_indet_face_vertices(cell, fkey):
    """Split all indeterminate vertices of a cell face.

    Parameters
    ----------
    cell : Mesh
        Cell as a mesh object.
    fkey : hashable
        Identifier of the face.

    Returns
    -------
    Cell
        Updated cell.

    Notes
    -----
    An indeterminate vertex is defined as a vertex with degree or valency of 3 or greater.

    """
    egi = mesh_dual(cell, EGI)

    f_vkeys = cell.face_vertices(fkey)
    egi_nbr_vkeys = egi.vertex_neighbors(fkey)

    for egi_fkey in f_vkeys:

        fkeys = egi.face_vertices(egi_fkey)
        i = fkeys.index(fkey)
        fkeys = fkeys[i:] + fkeys[:i]

        egi_face_vertices = [key for key in fkeys if key not in egi_nbr_vkeys + [fkey]]

        vkey_del = egi_fkey
        x, y, z = cell.vertex_coordinates(vkey_del)

        for vkey in egi_face_vertices:

            f, g = egi.mesh_split_face(vkey_del, fkey, vkey)

            cell.delete_vertex(vkey_del)

            cell.add_vertex(key=f, x=x, y=y, z=z)
            cell.add_vertex(key=g, x=x, y=y, z=z)

            for new_fkey in fkeys:
                new_vkeys = egi.vertex_faces(new_fkey, ordered=True)
                cell.add_face(new_vkeys, fkey=new_fkey)
            vkey_del = g

    return cell


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cell edges
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def cell_collapse_short_edge(cell, u, v, min_length=0.1):
    """Collapse short edges of a cell.

    Parameters
    ----------
    cell : Mesh
        Cell as a mesh object.
    u : hashable
        The key of the start vertex.
    v : hashable
        The key of the end vertex.
    min_length : float
        Minimum length of edges to be collapsed.

    Returns
    -------
    cell : Mesh
        Updated cell.

    """
    sp = cell.vertex_coordinates(u)
    ep = cell.vertex_coordinates(v)
    dist = distance_point_point(sp, ep)

    if dist < min_length:
        mp = midpoint_point_point(sp, ep)
        cell.vertex_update_xyz(u, mp)
        cell.vertex_update_xyz(v, mp)

    return cell


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   cell face operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def cell_cull_zero_faces(cell):
    pass


def cell_relocate_face(cell, fkey, xyz, normal):
    """Relocate the face of a mesh.

    Parameters
    ----------
    cell : Mesh
        Cell as a mesh object.
    fkey : hashable
        Identifier of the face.
    xyz : tuple
        xyz coordinates of the new target plane.
    normal : tuple
        Target normal vector.

    Returns
    -------
    cell : Mesh
        Updated cell.
    """

    cell_split_indet_face_vertices(cell, fkey)

    vkeys = cell.face_vertices(fkey)

    # new target plane for the face
    plane = (xyz, normal)

    # neighboring edges
    edges = {}
    for u in vkeys:
        for v in cell.vertex_neighbors(u):
            if v not in vkeys:
                edges[u] = v

    for u in edges:
        line = cell.edge_coordinates(u, edges[u])
        it = intersection_line_plane(line, plane)
        cell.vertex_update_xyz(u, it, constrained=False)

    return cell


def cell_merge_coplanar_adjacent_faces(cell, tol=0.001):

    initial_faces = [key for key in cell.face]
    current_faces = [key for key in cell.face]

    for fkey in initial_faces:

        if fkey in current_faces:
            normal = cell.face_normal(fkey)

            faces_to_delete = []

            for nbr_fkey in cell.face_neighbours(fkey):
                nbr_normal = cell.face_normal(nbr_fkey)
                dot = dot_vectors(normal, nbr_normal)
                if 1 - dot < tol:
                    faces_to_delete.append(nbr_fkey)

            if faces_to_delete:
                new_halfedges = cell.face_halfedges(fkey)
                for del_fkey in faces_to_delete:
                    for u, v in cell.face_halfedges(del_fkey):
                        if (v, u) in new_halfedges:
                            new_halfedges.remove((v, u))
                        else:
                            new_halfedges.append((u, v))

                new_face = list(new_halfedges[0])

                for i in range(1, len(new_halfedges)):
                    u, v = new_halfedges[i]
                    if u in new_face:
                        index = new_face.index(u)
                        new_face.insert(index + 1, v)
                    else:
                        index = new_face.index(v)
                        new_face.insert(index, u)

                for key in faces_to_delete:
                    cell.delete_face(key)
                    current_faces.remove(key)

                cell.add_face(vertices=new_face, fkey=fkey)

    return cell


def cell_face_subdivide_barycentric(cell, fkey):
    raise NotImplementedError


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import print_function

import io
from os import path

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


here = path.abspath(path.dirname(__file__))


def read(*names, **kwargs):
    return io.open(
        path.join(here, *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


long_description = read('README.md')
requirements = read('requirements.txt').split('\n')
optional_requirements = {}

setup(
    name='compas_3gs',
    version='0.5.2',
    description='A COMPAS package for 3D Graphic Statics',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/BlockResearchGroup/compas_3gs',
    author='Juney Lee',
    author_email='juney.lee@arch.ethz.ch',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords=['architecture', 'engineering'],
    project_urls={},
    packages=['compas_3gs', ],
    package_dir={'': 'src'},
    package_data={},
    data_files=[],
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    python_requires='>=2.7',
    extras_require=optional_requirements,
    entry_points={},
    ext_modules=[]
)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino
from compas_rhino.ui import CommandMenu
from compas.datastructures import Network
from compas.datastructures import Mesh
from compas.datastructures import mesh_subdivide_catmullclark
from compas.utilities import geometric_key
from compas.datastructures import meshes_join
from compas.geometry import closest_point_in_cloud
from compas.geometry import convex_hull
from compas.geometry import add_vectors
from compas.topology import unify_cycles
from compas.utilities import flatten
from compas_skeleton.datastructure import Skeleton3D_Node
from compas_skeleton.datastructure import Skeleton3D
from compas_rhino.artists import MeshArtist


__commandname__ = "TGS_skeleton"


def create_sk3_quad(lines, joint_width=1, leaf_width=1, joint_length=0.4):

    def get_convex_hull_mesh(points):
        faces = convex_hull(points)
        vertices = list(set(flatten(faces)))

        i_index = {i: index for index, i in enumerate(vertices)}
        vertices = [points[index] for index in vertices]
        faces = [[i_index[i] for i in face] for face in faces]
        faces = unify_cycles(vertices, faces)

        mesh = Mesh.from_vertices_and_faces(vertices, faces)

        return mesh

    def create_networks():
        networks = {}
        descendent_tree = {}
        for u in joints:
            global_local = {}
            lines = []
            nbrs = network_global.neighbors(u)
            start_pt = network_global.node_coordinates(u)

            for v in nbrs:
                end_pt = network_global.edge_point(u, v, t=joint_length)
                lines.append([start_pt, end_pt])

            network_local = Network.from_lines(lines)
            key_local = list(set(list(network_local.nodes())) - set(network_local.leaves()))[0]
            global_local.update({u: key_local})

            gkeys_global_network = [geometric_key(line[1]) for line in lines]
            gkeys_local_network = [
                geometric_key(network_local.node_coordinates(key))
                for key in network_local.leaves()
            ]

            for i, key_global in enumerate(nbrs):
                gkey_global = gkeys_global_network[i]
                index_local = gkeys_local_network.index(gkey_global)
                key_local = network_local.leaves()[index_local]
                global_local.update({key_global: key_local})

            descendent_tree.update({u: global_local})
            networks.update({u: network_local})

        return networks, descendent_tree

    def create_sk3_branch(u, v):
        def find_vertices(u, v):
            sk3_joint_u = sk3_joints[u]

            # inside of network_u, find vertices on the verge
            leaf_u = descendent_tree[u][v]  # this is network local key, not convexhull mesh
            leaf_u = sk3_joint_u.network_convexhull[leaf_u]
            nbrs = sk3_joint_u.convexhull_mesh.vertex_neighbors(leaf_u, ordered=True)
            keys = [sk3_joint_u.descendent_tree[leaf_u][nbr]['lp'] for nbr in nbrs]
            points = [sk3_joint_u.vertex_coordinates(key) for key in keys]

            return points

        if u in joints and v in joints:
            # its an internal edge
            points_u = find_vertices(u, v)
            points_v = find_vertices(v, u)

            if len(points_u) != len(points_v):
                mesh = get_convex_hull_mesh(points_u + points_v)
            else:
                points_v = points_v[::-1]
                index = closest_point_in_cloud(points_u[0], points_v)[2]
                points_v = points_v[index:] + points_v[:index]

                vertices = points_u + points_v
                faces = []
                n = len(points_u)
                for i in range(n):
                    faces.append(
                        [i, (i + 1) % n, (i + 1) % n + n, i + n]
                    )

                mesh = Mesh.from_vertices_and_faces(vertices, faces)

        else:
            if u in leafs:
                leaf, joint = u, v
            elif v in leafs:
                leaf, joint = v, u

            points_joint = find_vertices(joint, leaf)
            # network = networks[joint]

            # u_local = descendent_tree[joint][joint]
            # v_local = descendent_tree[joint][leaf]

            vec = [
                i * (1 - joint_length) for i in network_global.edge_vector(joint, leaf)
            ]
            points_leaf = [add_vectors(pt, vec) for pt in points_joint]

            vertices = points_joint + points_leaf
            faces = []
            n = len(points_joint)
            for i in range(n):
                faces.append(
                    [i, (i + 1) % n, (i + 1) % n + n, i + n]
                )

            mesh = Mesh.from_vertices_and_faces(vertices, faces)

        return mesh

    def create_sk3_branches():

        return [create_sk3_branch(u, v) for u, v in network_global.edges()]

    def create_sk3_joints(networks):
        sk3_joints = {}
        for u in networks:
            network = networks[u]
            sk3_joint = Skeleton3D_Node.from_network(network)
            sk3_joint.joint_width = joint_width
            sk3_joint.leaf_width = leaf_width
            sk3_joint.update_vertices_location()
            sk3_joints.update({u: sk3_joint})

        return sk3_joints

    def draw_mesh_faces(mesh):
        fkeys_nodraw = [fkey for fkey in mesh.faces() if mesh.face_area(fkey) <= 0]
        fkeys = list(set(list(mesh.faces())) - set(fkeys_nodraw))

        artist = MeshArtist(mesh)
        artist.layer = '3GS::Skeleton'
        artist.draw_faces(faces=fkeys, join_faces=True)

    network_global = Network.from_lines(lines)

    joints = []
    leafs = []
    for key in network_global.node:
        if network_global.is_leaf(key):
            leafs.append(key)
        else:
            joints.append(key)

    networks, descendent_tree = create_networks()
    sk3_joints = create_sk3_joints(networks)
    sk3_branches = create_sk3_branches()

    mesh = meshes_join(sk3_joints.values() + sk3_branches)
    draw_mesh_faces(mesh)

    return mesh


def create_sk3_exo(lines, branch_radius=1, node_radius_fac=1, segments=4):
    sk3 = Skeleton3D.from_skeleton_lines(lines)
    sk3.section_seg = segments
    sk3.branch_radius = branch_radius
    sk3.node_radius_fac = node_radius_fac
    sk3.generate_mesh()
    sk3.merge_triangles()

    artist = MeshArtist(sk3)
    artist.layer = '3GS::Skeleton'
    guid = artist.draw_faces(join_faces=True)
    artist.redraw()

    return sk3, guid[0]


def subdivide(mesh, guid):
    compas_rhino.delete_object(guid)
    mesh = mesh_subdivide_catmullclark(mesh, k=1)

    artist = MeshArtist(mesh)
    artist.layer = '3GS::Skeleton'
    guid = artist.draw_faces(join_faces=True)
    artist.redraw()

    return mesh, guid[0]


config = {
    "name": "skeleton_type",
    "message": "SkeletonType",
    "options": [
        {
            "name": "Exoskeleton",
            "message": "Exoskeleton",
            "action": create_sk3_exo
        },
        {
            "name": "QaudSkeleton",
            "message": "QaudSkeleton",
            "action": create_sk3_quad
        },
    ]
}

config_modify = {
    "name": "modify",
    "message": "Modify",
    "options": [
        {
            "name": "Finish",
            "message": "Finish",
            "action": None
        },
        {
            "name": "Subdivide",
            "message": "Subdivide",
            "action": subdivide
        },
    ]
}


def RunCommand(is_interactive):

    if '3GS' not in sc.sticky:
        compas_rhino.display_message('3GS has not been initialised yet.')
        return

    scene = sc.sticky['3GS']['scene']

    # get ForceVolMeshObject from scene
    objects = scene.find_by_name('form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    # --------------------------------------------------------------------------

    # draw skeleton from form ...

    menu = CommandMenu(config)
    action = menu.select_action()

    lines = [form.network.edge_coordinates(u, v) for u, v in form.network.edges()]

    if action['name'] == 'Exoskeleton':
        branch_radius = compas_rhino.rs.GetReal('branch radius:')
        node_radius_fac = compas_rhino.rs.GetReal('node size factor:', number=1, minimum=1)
        # segments = compas_rhino.rs.GetInteger('profile segments?', number=4, minimum=3)
        kwargs = {
            'lines': lines,
            'branch_radius': branch_radius,
            'node_radius_fac': node_radius_fac,
            # 'segments': segments
        }

        sk3, guid = action['action'](**kwargs)

    else:
        raise NotImplementedError
        # joint_width = compas_rhino.rs.GetReal('joint node widths:')
        # leaf_width = compas_rhino.rs.GetReal('leaf node widths:')
        # kwargs = {
        #     'lines': lines,
        #     'joint_width': joint_width,
        #     'leaf_width': leaf_width
        # }
        # action['action'](**kwargs)

    while True:
        menu = CommandMenu(config_modify)
        action = menu.select_action()
        if not action:
            return

        if action['name'] == 'Finish':
            break

        sk3, guid = action['action'](sk3, guid)

    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":

    RunCommand(True)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import json
from ast import literal_eval

import scriptcontext as sc

import compas_rhino
from compas_rhino.forms import TextForm

from compas_3gs.diagrams import FormNetwork
from compas_3gs.diagrams import ForceVolMesh


__all__ = [
    "is_valid_file",
    "select_filepath_open",
    "select_filepath_save",
    "get_tgs",
    "get_scene",
    "get_proxy",
    "get_system",
    "select_nodes",
    "select_vertices",
    "select_edges",
    "select_faces",
    "tgs_undo",
    "relocate_formdiagram"

]


def match_nodes(diagram, keys):
    temp = compas_rhino.get_objects(name="{}.node.*".format(diagram.name))
    names = compas_rhino.get_object_names(temp)
    guids = []
    for guid, name in zip(temp, names):
        parts = name.split('.')
        key = literal_eval(parts[2])
        if key in keys:
            guids.append(guid)
    return guids


def match_vertices(diagram, keys):
    temp = compas_rhino.get_objects(name="{}.vertex.*".format(diagram.name))
    names = compas_rhino.get_object_names(temp)
    guids = []
    for guid, name in zip(temp, names):
        parts = name.split('.')
        key = literal_eval(parts[2])
        if key in keys:
            guids.append(guid)
    return guids


def match_edges(diagram, keys):
    temp = compas_rhino.get_objects(name="{}.edge.*".format(diagram.name))
    names = compas_rhino.get_object_names(temp)
    guids = []
    for guid, name in zip(temp, names):
        parts = name.split('.')[2].split('-')
        u = literal_eval(parts[0])
        v = literal_eval(parts[1])
        if (u, v) in keys or (v, u) in keys:
            guids.append(guid)
    return guids


def match_faces(diagram, keys):
    temp = compas_rhino.get_objects(name="{}.face.*".format(diagram.name))
    names = compas_rhino.get_object_names(temp)
    guids = []
    for guid, name in zip(temp, names):
        parts = name.split('.')
        key = literal_eval(parts[2])
        if key in keys:
            guids.append(guid)
    return guids


def select_nodes(diagram, keys):
    guids = match_nodes(diagram, keys)
    compas_rhino.rs.EnableRedraw(False)
    compas_rhino.rs.SelectObjects(guids)
    compas_rhino.rs.EnableRedraw(True)


def select_vertices(diagram, keys):
    guids = match_vertices(diagram, keys)
    compas_rhino.rs.EnableRedraw(False)
    compas_rhino.rs.SelectObjects(guids)
    compas_rhino.rs.EnableRedraw(True)


def select_edges(diagram, keys):
    guids = match_edges(diagram, keys)
    compas_rhino.rs.EnableRedraw(False)
    compas_rhino.rs.SelectObjects(guids)
    compas_rhino.rs.EnableRedraw(True)


def select_faces(diagram, keys):
    guids = match_faces(diagram, keys)
    compas_rhino.rs.EnableRedraw(False)
    compas_rhino.rs.SelectObjects(guids)
    compas_rhino.rs.EnableRedraw(True)


def is_valid_file(filepath, ext):
    """Is the selected path a valid file.

    Parameters
    ----------
    filepath
    """
    if not filepath:
        return False
    if not os.path.exists(filepath):
        return False
    if not os.path.isfile(filepath):
        return False
    if not filepath.endswith(".{}".format(ext)):
        return False
    return True


def select_filepath_open(root, ext):
    """Select a filepath for opening a session.

    Parameters
    ----------
    root : str
        Base directory from where the file selection is started.
        If no directory is provided, the parent folder of the current
        Rhino document will be used
    ext : str
        The type of file that can be openend.

    Returns
    -------
    tuple
        The parent directory.
        The file name.
    None
        If the procedure fails.

    Notes
    -----
    The file extension is only used to identify the type of session file.
    Regardless of the provided extension, the file contents should be in JSON format.

    """
    ext = ext.split('.')[-1]
    filepath = compas_rhino.select_file(folder=root, filter=ext)
    if not is_valid_file(filepath, ext):
        print("This is not a valid session file: {}".format(filepath))
        return
    return filepath


def select_filepath_save(root, ext):
    """Select a filepath for saving a session."""
    filepath = compas_rhino.rs.SaveFileName('save', filter=ext, folder=root)
    if not filepath:
        return
    if filepath.split('.')[-1] != ext:
        filepath = "%s.%s" % (filepath, ext)
    return filepath


def get_tgs():
    if "3GS" not in sc.sticky:
        form = TextForm('Initialise the plugin first!', '3GS')
        form.show()
        return None
    return sc.sticky["3GS"]


def get_scene():
    tgs = get_tgs()
    if tgs:
        return tgs['scene']


def get_proxy():
    if "3GS.proxy" not in sc.sticky:
        form = TextForm('Initialise the plugin first!', '3GS')
        form.show()
        return None
    return sc.sticky["3GS.proxy"]


def get_system():
    if "3GS.system" not in sc.sticky:
        form = TextForm('Initialise the plugin first!', '3GS')
        form.show()
        return None
    return sc.sticky["3GS.system"]


def save_session():
    scene = get_scene()
    session = {
        "data": {"form": None, "force": None},
        "settings": scene.settings,
    }
    form = scene.get('form')[0]
    if form:
        session['data']['form'] = form.datastructure.to_data()
    force = scene.get('force')[0]
    if force:
        session['data']['force'] = force.datastructure.to_data()
    return session


def load_session(session):
    print("loading session")
    scene = get_scene()
    scene.clear()
    if 'settings' in session:
        scene.settings = session['settings']
    if 'data' in session:
        data = session['data']
        if 'form' in data and data['form']:
            form = FormNetwork.from_data(data['form'])
            scene.add(form, name="form")

        if 'force' in data and data['force']:
            force = ForceVolMesh.from_data(data['force'])
            force.primal = form
            form.dual = force
            force.update_angle_deviations()
            scene.add(force, name="force")
    scene.update()


def record():
    session = json.loads(json.dumps(save_session()))
    sc.sticky["3GS.sessions"] = sc.sticky["3GS.sessions"][:sc.sticky["3GS.sessions.current"] + 1]
    sc.sticky["3GS.sessions"].append(session)
    if len(sc.sticky["3GS.sessions"]) > 10:
        sc.sticky["3GS.sessions"] = sc.sticky["3GS.sessions"][-10:]
    sc.sticky["3GS.sessions.current"] = len(sc.sticky["3GS.sessions"]) - 1


def undo(sender, e):
    if e.Tag == "undo":
        if sc.sticky["3GS.sessions.current"] - 1 < 0:
            print("no more recorded sessions to undo")
            return
        sc.sticky["3GS.sessions.current"] -= 1
        session = sc.sticky["3GS.sessions"][sc.sticky["3GS.sessions.current"]]
        load_session(session)
        e.Document.AddCustomUndoEvent("RV2 Redo", undo, "redo")
    if e.Tag == "redo":
        if sc.sticky["3GS.sessions.current"] + 1 >= len(sc.sticky["3GS.sessions"]):
            print("no more recorded sessions to redo")
            return
        sc.sticky["3GS.sessions.current"] += 1
        session = sc.sticky["3GS.sessions"][sc.sticky["3GS.sessions.current"]]
        load_session(session)
        e.Document.AddCustomUndoEvent("RV2 Redo", undo, "undo")
    print("current sessions:", sc.sticky["3GS.sessions.current"] + 1)
    print("total sessions:", len(sc.sticky["3GS.sessions"]))


def tgs_undo(command):
    def wrapper(*args, **kwargs):
        sc.doc.EndUndoRecord(sc.doc.CurrentUndoRecordSerialNumber)
        undoRecord = sc.doc.BeginUndoRecord("RV2 Undo")
        if undoRecord == 0:
            print("undo record did not start")
        else:
            print("Custom undo recording", undoRecord)

        if len(sc.sticky["3GS.sessions"]) == 0:
            sc.sticky["3GS.sessions.current"] = 0
            record()
        command(*args, **kwargs)
        record()
        sc.doc.AddCustomUndoEvent("3GS Undo", undo, "undo")
        if undoRecord > 0:
            sc.doc.EndUndoRecord(undoRecord)
    return wrapper


def relocate_formdiagram(force, form):

    bbox_force = force.bounding_box()
    bbox_form = form.bounding_box()
    xmin_force, xmax_force = bbox_force[0][0], bbox_force[1][0]
    xmin_form, _ = bbox_form[0][0], bbox_form[1][0]
    ymin_force, ymax_force = bbox_force[0][1], bbox_force[3][1]
    ymin_form, ymax_form = bbox_form[0][1], bbox_form[3][1]
    y_force = ymin_force + 0.5 * (ymax_force - ymin_force)
    y_form = ymin_form + 0.5 * (ymax_form - ymin_form)
    dx = 1.5 * (xmax_force - xmin_force) + (xmin_force - xmin_form)
    dy = y_force - y_form

    return (dx, dy, 0)



from compas.geometry import add_vectors
from compas.geometry import subtract_vectors
from compas.geometry import scale_vector

from itertools import groupby

from compas.datastructures import Mesh

from compas.geometry import convex_hull

from compas_3gs.utilities import datastructure_centroid

from compas_3gs.operations import cell_merge_coplanar_adjacent_faces


__all__ = ['point_reflection',

           'volmesh_vertex_lift',
           'volmesh_vertex_merge',

           'volmesh_halfface_pinch',
           'volmesh_halfface_extrude',
           'volmesh_halfface_subdivide',
           'volmesh_merge_adjacent_halffaces',

           'volmesh_cell_split_vertex',
           'volmesh_cell_subdivide_barycentric']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh global operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def point_reflection(datastructure):
    """Inverts a datastructure through its centroid.

    Parameters
    ----------
    datastructure
        A network, mesh or volmesh object.

    Returns
    -------
        The datastructure with new vertex coordinates.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Point_reflection

    """
    center = datastructure_centroid(datastructure)

    for vkey in datastructure.vertex:
        xyz = datastructure.vertex_coordinates(vkey)
        vector = scale_vector(subtract_vectors(center, xyz), 2)
        new_xyz = add_vectors(xyz, vector)
        datastructure.vertex_update_xyz(vkey, new_xyz, constrained=False)


def reverse_cycle_directions(volmesh):
    raise NotImplementedError


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh vertex operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def volmesh_vertex_lift(volmesh, vkey, target_xyz, hfkeys):
    """Duplicates and lifts a vertex, then creates cells using the vertex halffaces and the duplicated vertex.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    vkey : int
        The key of the vertex to lift.
    hfkeys : list
        List of halfface keys to create new cells from.
    xyz : tuple
        Target xyz coordinates of the lifted vertex.

    Notes
    -----
    The lifting vertex must be on the boundary of the volmesh.

    See Also
    --------
    * :func:`compas_3gs.operations.face_pinch`

    """

    # check if vertex is interior ----------------------------------------------
    if not volmesh.is_vertex_on_boundary(vkey):
        raise Exception('This vertex is interior.')

    # add new, lifted vertex ---------------------------------------------------
    x, y, z = target_xyz
    w = volmesh.add_vertex(x=x, y=y, z=z)

    # add cells ----------------------------------------------------------------
    for hfkey in hfkeys:
        halffaces = [volmesh.halfface_vertices(hfkey)[::-1]]
        for u, v in volmesh.halfface_halfedges(hfkey):
            halffaces.append([u, v, w])
        volmesh.add_cell(halffaces)

    # --------------------------------------------------------------------------
    return volmesh


def volmesh_vertex_merge(volmesh, vkeys, target_xyz):
    """Merges specified vertices to a single vertex.

    Parameters
    ----------
    volmesh : VolMesh
        A volmesh object representing a polyhedral force diagram.
    vkeys : list
        List of vertices to merge.
    target_xyz : tuple
        Target xyz coordinates of the vertices to be merged.

    Notes
    -----
    As implemented, the merging vertices (in any combination) must be a continuous chain of halfedges of a halfface.
    For example, vertices 1 and 3 cannot be merged.

    0 ------ 1
    |        |
    |        |
    |        |
    3 ------ 2

    """

    # inspect vertices ---------------------------------------------------------
    hf_vkeys = {}
    for vkey in vkeys:
        for hfkey in volmesh.vertex_halffaces(vkey):
            if hfkey not in hf_vkeys:
                hf_vkeys[hfkey] = []
            hf_vkeys[hfkey].append(vkey)

    for hfkey in hf_vkeys:
        all_vkeys = volmesh.halfface[hfkey]
        test_vkeys = set(hf_vkeys[hfkey])

        if len(test_vkeys) == 1:
            continue

        if len(all_vkeys) - len(test_vkeys) <= 1:
            raise Exception("Illegal merge!")

        for vkey in test_vkeys:
            anc = volmesh.halfface_vertex_ancestor(hfkey, vkey)
            des = volmesh.halfface_vertex_descendent(hfkey, vkey)
            if anc not in test_vkeys and des not in test_vkeys:
                raise Exception("Illegal merge!")

    # get vertex cells ---------------------------------------------------------
    new_vkey = volmesh._get_vertex_key(None)
    vertices = {new_vkey: target_xyz}

    cell_vkeys = {}
    for vkey in vkeys:
        for ckey in volmesh.vertex_cells(vkey):
            if ckey not in cell_vkeys:
                cell_vkeys[ckey] = []
            cell_vkeys[ckey].append(vkey)

    # construct new cell halffaces ---------------------------------------------
    cells = {}
    for ckey in cell_vkeys:
        cell_halffaces = {}
        vkeys = cell_vkeys[ckey]

        for hfkey in volmesh.cell_halffaces(ckey):
            halfface = [vkey for vkey in volmesh.halfface[hfkey]]
            for index, item in enumerate(halfface):
                if item in vkeys:
                    halfface[index] = new_vkey
            halfface = [key[0] for key in groupby(halfface)]
            if len(set(halfface)) > 2:
                cell_halffaces[hfkey] = halfface
        cells[ckey] = cell_halffaces
        for c_vkey in volmesh.cell_vertices(ckey):
            vertex_ckeys = volmesh.vertex_cells(c_vkey)
            if len(vertex_ckeys) == 1:
                if c_vkey not in vkeys:
                    vertices[c_vkey] = volmesh.vertex_coordinates(c_vkey)
        volmesh.delete_cell(ckey)

    # add vertices -------------------------------------------------------------
    for vkey in vertices:
        x, y, z = vertices[vkey]
        volmesh.add_vertex(vkey=vkey, x=x, y=y, z=z)

    # add halffaces and cells --------------------------------------------------
    for ckey in cells:
        volmesh.cell[ckey] = {}
        for hfkey in cells[ckey]:
            vertices = cells[ckey][hfkey]
            volmesh.add_halfface(vertices, fkey=hfkey)
            vertices = volmesh.halfface[hfkey]
            for i in range(-2, len(vertices) - 2):
                u = vertices[i]
                v = vertices[i + 1]
                w = vertices[i + 2]
                if u not in volmesh.cell[ckey]:
                    volmesh.cell[ckey][u] = {}
                volmesh.cell[ckey][u][v] = hfkey
                volmesh.plane[u][v][w] = ckey

    # --------------------------------------------------------------------------
    return volmesh


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh edge operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


# def _volmesh_cull_duplicate_edges(volmesh):
#     new_edges = {vkey: {} for vkey in volmesh.vertex}
#     seen      = set()
#     for u, v in volmesh.edges():
#         edge = frozenset([u, v])
#         if edge not in seen:
#             new_edges[u][v] = {}
#             seen.add(edge)
#     volmesh.edge = new_edges


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh halfface operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def volmesh_halfface_extrude(volmesh, hfkey):
    raise NotImplementedError


def volmesh_halfface_subdivide(volmesh, hfkey):
    raise NotImplementedError


def volmesh_halfface_pinch(volmesh, hfkey, xyz):
    """

    """

    x, y, z = xyz
    w = volmesh.add_vertex(x=x, y=y, z=z)
    hf_vkeys = volmesh.halfface_vertices(hfkey)
    cell_halffaces = [hf_vkeys[::-1]]
    halfedges = volmesh.halfface_halfedges(hfkey)
    for u, v in halfedges:
        cell_halffaces.append([u, v, w])
    volmesh.add_cell(cell_halffaces)
    return volmesh


def volmesh_merge_adjacent_halffaces(volmesh, hfkeys):

    # check halffaces ----------------------------------------------------------
    for hfkey in hfkeys:
        if not volmesh.is_halfface_on_boundary(hfkey):
            raise ValueError('Halfface {} is interior.'.format(hfkey))
    if not _are_halffaces_chained(volmesh, hfkeys):
        raise ValueError('These halffaces are not chained.')
    # --------------------------------------------------------------------------
    halffaces = [volmesh.halfface[hfkey] for hfkey in hfkeys]
    # --------------------------------------------------------------------------
    vkeys = set()
    for hfkey in hfkeys:
        for key in volmesh.halfface_vertices(hfkey):
            vkeys.add(key)

    vkeys = list(vkeys)
    points = [volmesh.vertex_coordinates(vkey) for vkey in vkeys]

    faces_by_index = convex_hull(points)
    faces_by_vkeys = []
    for face in faces_by_index:
        faces_by_vkeys.append([vkeys[index] for index in face])

    # make temp cell mesh ------------------------------------------------------
    cell = Mesh()
    for i in range(len(vkeys)):
        key = vkeys[i]
        x, y, z = points[i]
        cell.add_vertex(key=key, x=x, y=y, z=z)
    for face in faces_by_vkeys:
        cell.add_face(face)

    # merge coplanar faces -----------------------------------------------------
    cell_merge_coplanar_adjacent_faces(cell)

    # get correct direction of faces -------------------------------------------
    faces = [cell.face[fkey] for fkey in cell.face]
    if halffaces[0] in faces:
        new_faces = [face[::-1] for face in faces]
    else:
        new_faces = faces

    volmesh.add_cell(new_faces)

    return volmesh


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh cell operations
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def volmesh_cell_split_vertex(self, vkey, hfkey1, hfkey2):

    if len(self.cell) > 1:
        raise ValueError('This is a multi-cell volmesh.')

    ckey = self.cell.keys()[0]
    halffaces = self.cell_vertex_halffaces(ckey, vkey)
    i = halffaces.index(hfkey1)
    j = halffaces.index(hfkey2)

    if i + 1 == j or j + 1 == i:
        raise ValueError('The two halffaces are adjacent.')

    egi = self.c_data[ckey]['egi']
    hfkeys = egi.face_vertices(vkey)

    f, g = egi.mesh_split_face(vkey, hfkey1, hfkey2)

    x, y, z = self.vertex_coordinates(vkey)
    self.add_vertex(key=f, x=x, y=y, z=z)
    self.add_vertex(key=g, x=x, y=y, z=z)

    # new halffaces ------------------------------------------------------------
    for hfkey in hfkeys:
        new_vkeys = egi.vertex_faces(hfkey, ordered=True)
        self.add_halfface(new_vkeys[::-1], fkey=hfkey)

    self.cell_vertex_delete(vkey)

    print(egi.face)


def volmesh_cell_subdivide_barycentric(volmesh, ckey):
    """Subdivide a cell of a volmesh.

    """
    new_ckeys = []
    x, y, z = volmesh.cell_center(ckey)
    w = volmesh.add_vertex(x=x, y=y, z=z)

    halffaces = volmesh.cell_halffaces(ckey)

    for hfkey in halffaces:
        cell_halffaces = [volmesh.halfface_vertices(hfkey)]

        halfedges = volmesh.halfface_halfedges(hfkey)

        for u, v in halfedges:
            cell_halffaces.append([w, v, u])

        volmesh.delete_halfface(hfkey)
        new_ckeys.append(volmesh.add_cell(cell_halffaces))

    del volmesh.cell[ckey]

    return new_ckeys


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   volmesh operation helpers
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def _are_halffaces_chained(volmesh, hfkeys):
    chained = set()

    for hfkey_1 in hfkeys:
        nbrs = []

        for u, v in volmesh.halfface_halfedges(hfkey_1):
            nbrs += volmesh.edge_halffaces(u, v)

        for hfkey_2 in hfkeys:
            if hfkey_2 != hfkey_1 and hfkey_2 in nbrs:
                chained.add(hfkey_1)
                chained.add(hfkey_2)

    if len(hfkeys) == 1 or len(chained) == len(hfkeys):
        return True

    return False


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import dot_vectors

from compas.utilities import i_to_rgb
from compas.utilities import i_to_blue
from compas.utilities import i_to_green
from compas.utilities import i_to_red

from compas_3gs.utilities.topology import pair_uv_to_hf


__all__ = ['get_index_colordict',
           'valuedict_to_colordict',
           'compare_initial_current',

           'get_force_mags',
           'get_force_colors_uv',
           'get_force_colors_hf',

           'print_result']


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   general
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def get_index_colordict(key_list, color_scheme=i_to_rgb):
    """Convert a list of keys to a dictionary of key-color pairs based on a color gradient scheme.

    Parameters
    ----------
    key_list : list
        List of keys.
    color_scheme
        Desired color gradient scheme (i.e. i_to_rgb, i_to_red, i_to_blue, etc.)

    Returns
    -------
    c_dict : dict
        Dictionary of key-color pairs, from lowest to highest based on a color gradient scheme.

    """
    c_dict = {}

    if key_list:

        for index, key in enumerate(key_list):

            if len(key_list) == 1:
                value = 1
            else:
                value = float(index) / (len(key_list) - 1)

            color = color_scheme(value)
            c_dict[key] = color

    return c_dict


def valuedict_to_colordict(value_dict, color_scheme=i_to_rgb):
    """Convert a dictionary of key-value pairs to a dictionary of key-color pairs.

    Parameters
    ----------
    value_dict : dict
        Dictionary of key-value pairs.
    color_scheme
        Desired color gradient scheme (i.e. i_to_rgb, i_to_red, i_to_blue, etc.)

    Returns
    -------
    c_dict : dict
        Dictionary of key-color pairs, from lowest to highest based on a color gradient scheme.

    """
    c_dict = {}

    lb = min(value_dict.values())
    ub = max(value_dict.values())

    for key in value_dict:

        value = (value_dict[key] - lb) / (ub - lb)
        color = color_scheme(value)
        c_dict[key] = color

    return c_dict


def compare_initial_current(current_value_dict,
                            initial_value_dict,
                            color_scheme=i_to_rgb):
    """

    """
    color_dict = {}

    for key in current_value_dict:

        current = current_value_dict[key]
        initial = initial_value_dict[key]

        if current > 0.01:

            if initial < 0.01:
                value = 0
            else:
                value = current / initial

            color_dict[key] = color_scheme(value)

    return color_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   force related
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def get_force_mags(volmesh, network):
    """Returns a dictionary of (u,v)-magnitude pairs.
    Negative magnitude means compression, while positive magnitude means tension.
    """

    uv_hf_dict = pair_uv_to_hf(network, volmesh)

    mags = {}

    for u, v in network.edges():
        hfkey = uv_hf_dict[(u, v)]
        edge_vector = network.edge_vector(u, v)
        face_normal = volmesh.halfface_normal(hfkey)
        face_area = volmesh.halfface_area(hfkey)
        dot = dot_vectors(face_normal, edge_vector)

        if dot < 0:
            factor = -1

        if dot > 0:
            factor = 1

        force = face_area * factor * volmesh.attributes['convention']

        mags[(u, v)] = force

    return mags


def get_force_colors_uv(volmesh,
                        network,
                        gradient=False,
                        tol=0.001):
    """Returns a dictionary of (u,v)-color pairs.
    Blue means compression, and red means tension.
    """
    c_dict = {}
    f_dict = get_force_mags(volmesh, network)
    f_range = sorted(f_dict.values())

    c_forces = [x for x in f_range if x < 0]
    t_forces = [x for x in f_range if x > 0]

    for edge in f_dict:
        force = f_dict[edge]
        if force < 0:  # if compression
            color = (0, 0, 255)
            if gradient:
                min_c = abs(c_forces[-1])
                max_c = abs(c_forces[0])
                if min_c != max_c:
                    color = i_to_blue((abs(force) - min_c) / (max_c - min_c))

        if force > 0:  # if tension
            color = (255, 0, 0)
            if gradient:
                min_t = t_forces[0]
                max_t = t_forces[-1]
                if min_t != max_t:
                    color = i_to_red((force - min_t) / (max_t - min_t))

        if abs(force) < tol:  # if close to zero
            color = (255, 255, 255)

        c_dict[edge] = color

    return c_dict


def get_force_colors_hf(volmesh,
                        network,
                        uv_c_dict=None,
                        gradient=False,
                        boundary=False,
                        tol=0.001):
    """Returns a dictionary of hfkey-color pairs.
    """
    uv_hf_dict = pair_uv_to_hf(network, volmesh)

    uv_c_dict = uv_c_dict or get_force_colors_uv(volmesh,
                                                 network,
                                                 gradient=gradient)

    # interior halffaces -------------------------------------------------------
    hf_c_dict = {}
    for uv in uv_c_dict:
        u_hfkey = uv_hf_dict[uv]
        v_hfkey = volmesh.halfface_opposite_halfface(u_hfkey)

        hf_c_dict[u_hfkey] = uv_c_dict[uv]
        hf_c_dict[v_hfkey] = uv_c_dict[uv]

    # boundary halffaces -------------------------------------------------------
    if boundary:
        b_hfkeys = volmesh.halffaces_on_boundaries()
        b_hf_areas = {hfkey: volmesh.halfface_area(hfkey) for hfkey in b_hfkeys}
        b_hf_c_dict = valuedict_to_colordict(b_hf_areas,
                                             color_scheme=i_to_green)
        for hfkey in b_hfkeys:
            hf_color = (0, 255, 0)

            if gradient:
                hf_color = b_hf_c_dict[hfkey]
            hf_c_dict[hfkey] = hf_color

    return hf_c_dict


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   other
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


def print_result(name, k, deviation):
    """Prints the result of an algorithm.

    Parameters
    ----------
    name : string
        name of the algorithm
    k : int
        number of iterations
    deviation : float
        deviation

    """
    name = str(name)

    print('===================================================================')
    print('')
    print(name, 'ended after', k, 'iterations.')
    print('')
    print('Max deviation :', deviation)
    print('')
    print('===================================================================')


# ******************************************************************************
# ******************************************************************************
# ******************************************************************************
#
#   Main
#
# ******************************************************************************
# ******************************************************************************
# ******************************************************************************


if __name__ == '__main__':
    pass



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.utilities import i_to_rgb

import compas_rhino

from compas_3gs.rhino.objects.volmeshobject import VolMeshObject

from compas_3gs.utilities import volmesh_face_flatness

from compas_3gs.utilities import get_force_colors_hf


__all__ = ['ForceVolMeshObject']


class ForceVolMeshObject(VolMeshObject):
    """A force volmesh object represents a polyhedral force diagram in the Rhino view.
    """

    SETTINGS = {
        'layer': "3GS::ForceDiagram",

        '_is.valid': False,

        'show.vertices': True,
        'show.edges': True,
        'show.faces': True,
        'show.vertexlabels': False,
        'show.facelabels': False,
        'show.celllabels': False,

        'color.invalid': (100, 255, 100),

        'color.vertices': (0, 0, 0),
        'color.vertexlabels': (255, 255, 255),
        'color.vertices:is_fixed': (0, 0, 255),

        'color.edges': (50, 50, 50),

        'color.faces': (200, 200, 200),
        'color.facelabels': (200, 200, 200),

        'color.celllabels': (0, 0, 255)
    }

    def __init__(self, diagram, *args, **kwargs):
        super(ForceVolMeshObject, self).__init__(diagram, *args, **kwargs)
        self.settings.update(ForceVolMeshObject.SETTINGS)
        settings = kwargs.get('settings') or {}
        if settings:
            self.settings.update(settings)

    def check_eq(self):
        ftol = self.scene.settings['3GS']['tol.flatness']
        fmax = max(volmesh_face_flatness(self.diagram).values())

        atol = self.scene.settings['3GS']['tol.angles']
        halffaces = list(self.diagram.faces())
        amax = max(self.diagram.faces_attribute('_a', faces=halffaces))

        if fmax > ftol or amax > atol:
            self.settings['_is.valid'] = False
        if fmax < ftol and amax < atol:
            self.settings['_is.valid'] = True

    # @property
    # def vertex_xyz(self):
    #     """dict : The view coordinates of the mesh object."""
    #     origin = Point(0, 0, 0)
    #     if self.anchor is not None:
    #         xyz = self.volmesh.vertex_attributes(self.anchor, 'xyz')
    #         point = Point(* xyz)
    #         T1 = Translation.from_vector(origin - point)
    #         S = Scale.from_factors([self.scale] * 3)
    #         R = Rotation.from_euler_angles(self.rotation)
    #         T2 = Translation.from_vector(self.location)
    #         X = T2 * R * S * T1
    #     else:
    #         S = Scale.from_factors([self.scale] * 3)
    #         R = Rotation.from_euler_angles(self.rotation)
    #         T = Translation.from_vector(self.location)
    #         X = T * R * S
    #     volmesh = self.volmesh.transformed(X)
    #     vertex_xyz = {vertex: volmesh.vertex_attributes(vertex, 'xy') + [0.0] for vertex in volmesh.vertices()}
    #     return vertex_xyz
    # # --------------------------------------------------------------------------
    # #   attributes
    # # --------------------------------------------------------------------------

    # def update_attributes(self):
    #     """Update the attributes of the data structure through a Rhino dialog.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     return compas_rhino.update_settings(self.datastructure.attributes)

    # def update_vertices_attributes(self, keys, names=None):
    #     """Update the attributes of selected vertices.

    #     Parameters
    #     ----------
    #     keys : list
    #         The identifiers of the vertices of which the attributes should be updated.
    #     names : list, optional
    #         The names of the attributes that should be updated.
    #         Default is ``None``, in which case all attributes are updated.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     if keys:
    #         compas_rhino.rs.UnselectAllObjects()
    #         select_vertices(self.datastructure, keys)
    #         return mesh_update_vertex_attributes(self.datastructure, keys, names)

    # def update_edges_attributes(self, keys, names=None):
    #     """Update the attributes of selected edges.

    #     Parameters
    #     ----------
    #     keys : list
    #         The identifiers of the edges of which the attributes should be updated.
    #     names : list, optional
    #         The names of the attributes that should be updated.
    #         Default is ``None``, in which case all attributes are updated.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     if keys:
    #         compas_rhino.rs.UnselectAllObjects()
    #         select_edges(self.datastructure, keys)
    #         return mesh_update_edge_attributes(self.datastructure, keys, names)

    # def update_faces_attributes(self, keys, names=None):
    #     """Update the attributes of selected faces.

    #     Parameters
    #     ----------
    #     keys : list
    #         The identifiers of the faces of which the attributes should be updated.
    #     names : list, optional
    #         The names of the attributes that should be updated.
    #         Default is ``None``, in which case all attributes are updated.

    #     Returns
    #     -------
    #     bool
    #         True if the update was successful.
    #         False otherwise.
    #     """
    #     if keys:
    #         compas_rhino.rs.UnselectAllObjects()
    #         select_faces(self.datastructure, keys)
    #         return mesh_update_face_attributes(self.datastructure, keys, names)

    # --------------------------------------------------------------------------
    #   drawing
    # --------------------------------------------------------------------------

    def draw(self):
        """Draw the objects representing the force diagram.
        """
        layer = self.settings["layer"]
        self.artist.layer = layer
        self.artist.clear_layer()
        self.clear()
        if not self.visible:
            return

        self.artist.vertex_xyz = self.vertex_xyz

        # ======================================================================
        # Groups
        # ======================================================================

        group_vertices = "{}::vertices".format(layer)
        group_vertices_labels = "{}::vertices_labels".format(layer)
        group_edges = "{}::edges".format(layer)
        group_edges_labels = "{}::edges_labels".format(layer)
        group_halffaces = "{}::halffaces".format(layer)
        group_halffaces_labels = "{}::halffaces_labels".format(layer)
        group_cells_labels = "{}::cells_labels".format(layer)
        group_angles = "{}::angles".format(layer)

        if not compas_rhino.rs.IsGroup(group_vertices):
            compas_rhino.rs.AddGroup(group_vertices)

        if not compas_rhino.rs.IsGroup(group_vertices_labels):
            compas_rhino.rs.AddGroup(group_vertices_labels)

        if not compas_rhino.rs.IsGroup(group_edges):
            compas_rhino.rs.AddGroup(group_edges)

        if not compas_rhino.rs.IsGroup(group_edges_labels):
            compas_rhino.rs.AddGroup(group_edges_labels)

        if not compas_rhino.rs.IsGroup(group_halffaces):
            compas_rhino.rs.AddGroup(group_halffaces)

        if not compas_rhino.rs.IsGroup(group_halffaces_labels):
            compas_rhino.rs.AddGroup(group_halffaces_labels)

        if not compas_rhino.rs.IsGroup(group_cells_labels):
            compas_rhino.rs.AddGroup(group_cells_labels)

        if not compas_rhino.rs.IsGroup(group_angles):
            compas_rhino.rs.AddGroup(group_angles)

        # ======================================================================
        # vertices
        # ======================================================================

        # vertices -------------------------------------------------------------
        vertices = list(self.diagram.vertices())
        color = {vertex: self.settings['color.vertices'] if self.settings['_is.valid'] else self.settings['color.invalid'] for vertex in vertices}
        for vertex in vertices:
            if self.diagram.vertex_attribute(vertex, 'is_fixed'):
                color[vertex] = self.settings['color.vertices:is_fixed']

        if self.settings["show.vertices"]:
            guids = self.artist.draw_vertices(color=color)
            self.guid_vertex = zip(guids, vertices)
            compas_rhino.rs.AddObjectsToGroup(guids, group_vertices)

            compas_rhino.rs.ShowGroup(group_vertices)
        else:
            compas_rhino.rs.HideGroup(group_vertices)

        # vertices labels ------------------------------------------------------
        if self.settings["show.vertexlabels"]:
            text = {vertex: index for index, vertex in enumerate(vertices)}

            guids = self.artist.draw_vertexlabels(text=text, color=color)
            self.guid_vertexlabel = zip(guids, vertices)
            compas_rhino.rs.AddObjectsToGroup(guids, group_vertices_labels)

            compas_rhino.rs.ShowGroup(group_vertices_labels)
        else:
            compas_rhino.rs.HideGroup(group_vertices_labels)

        # ======================================================================
        # edges
        # ======================================================================

        # edges ----------------------------------------------------------------
        edges = list(self.diagram.edges())
        color = {edge: self.settings['color.edges'] if self.settings['_is.valid'] else self.settings['color.invalid'] for edge in edges}

        guids = self.artist.draw_edges(edges, color)
        self.guid_edge = zip(guids, edges)
        compas_rhino.rs.AddObjectsToGroup(guids, group_edges)

        if self.settings["show.edges"] and self.settings['_is.valid']:
            compas_rhino.rs.ShowGroup(group_edges)
        else:
            compas_rhino.rs.HideGroup(group_edges)

        # # edge labels ----------------------------------------------------------
        # text = {edge: index for index, edge in enumerate(edges)}
        # guids = self.artist.draw_edgelabels(text=text, color=edges_labels_color)
        # self.guid_edgelabel = zip(guids, edges)
        # compas_rhino.rs.AddObjectsToGroup(guids, group_edges_labels)

        # if self.settings["show.edgelabels"]:
        #     compas_rhino.rs.ShowGroup(group_edges_labels)
        # else:
        #     compas_rhino.rs.HideGroup(group_edges_labels)

        # ======================================================================
        # halffaces
        # ======================================================================

        # halffaces ------------------------------------------------------------
        halffaces = list(self.diagram.faces())

        colors = {face: self.settings['color.faces'] for face in halffaces}

        if self.scene.settings['3GS']['show.forces']:
            colors = get_force_colors_hf(self.diagram,
                                         self.diagram.primal,
                                         gradient=True)
            halffaces = colors.keys()

        colordict = {face: colors[face] if self.settings['_is.valid'] else self.settings['color.invalid'] for face in halffaces}

        if self.settings['show.faces']:
            guids = self.artist.draw_faces(halffaces, colordict)
            self.guid_face = zip(guids, halffaces)

        # compas_rhino.rs.AddObjectsToGroup(guids, group_halffaces)

        # if self.settings['show.faces']:
        #     compas_rhino.rs.ShowGroup(group_halffaces)
        # else:
        #     compas_rhino.rs.HideGroup(group_halffaces)

        # halfface labels ------------------------------------------------------
        if self.settings["show.facelabels"] and self.settings['_is.valid']:
            text = {halfface: index for index, halfface in enumerate(halffaces)}
            guids = self.artist.draw_facelabels(text=text, color=color)
            self.guid_facelabel = zip(guids, halffaces)
            compas_rhino.rs.AddObjectsToGroup(guids, group_halffaces_labels)

            compas_rhino.rs.ShowGroup(group_halffaces_labels)
        else:
            compas_rhino.rs.HideGroup(group_halffaces_labels)

        # ======================================================================
        # cell labels
        # ======================================================================

        if self.settings["show.celllabels"] and self.settings['_is.valid']:
            cells = list(self.diagram.cells())
            cells_labels_color = {}
            cells_labels_color.update({cell: self.settings['color.celllabels'] for cell in cells})
            guids = self.artist.draw_celllabels(color=cells_labels_color)
            self.guid_celllabel = zip(guids, cells)
            compas_rhino.rs.AddObjectsToGroup(guids, group_cells_labels)

            compas_rhino.rs.ShowGroup(group_cells_labels)
        else:
            compas_rhino.rs.HideGroup(group_cells_labels)

        # ======================================================================
        # angle deviations
        # ======================================================================

        if self.scene and self.scene.settings['3GS']['show.angles']:
            tol = self.scene.settings['3GS']['tol.angles']
            halffaces = [halfface for halfface in self.diagram.faces() if not self.diagram.is_halfface_on_boundary(halfface)]
            angles = self.diagram.faces_attribute('_a', faces=halffaces)
            amin = min(angles)
            amax = max(angles)
            if (amax - amin)**2 > 0.001**2:
                text = {}
                color = {}
                for halfface, angle in zip(halffaces, angles):
                    if angle > tol:
                        text[halfface] = "{:.0f}".format(angle)
                        color[halfface] = i_to_rgb((angle - amin) / (amax - amin))
                if text:
                    guids = self.artist.draw_facelabels(text, color)
                    self.guid_facelabel = zip(guids, halffaces)
                    compas_rhino.rs.AddObjectsToGroup(guids, group_angles)
                    compas_rhino.rs.ShowGroup(group_angles)
        else:
            compas_rhino.rs.HideGroup(group_angles)

        self.redraw()



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from math import pi
from math import sqrt

import compas_rhino

from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import length_vector

from compas.utilities import i_to_green

from compas_3gs.rhino.artists.networkartist import NetworkArtist

from compas_3gs.utilities import valuedict_to_colordict
from compas_3gs.utilities import get_force_mags


__all__ = ['FormNetworkArtist']


class FormNetworkArtist(NetworkArtist):
    """Artist for visualizing force diagrams in the Rhino model space."""

    def __init__(self, form, layer=None):
        super(FormNetworkArtist, self).__init__(form, layer=layer)

    @property
    def node_xyz(self):
        """dict:
        The view coordinates of the network vertices.
        The view coordinates default to the actual network coordinates.
        """
        if not self._node_xyz:
            self._node_xyz = {node: self.diagram.node_attributes(node, 'xyz') + [0.0] for node in self.diagram.nodes()}
        return self._node_xyz

    @node_xyz.setter
    def node_xyz(self, node_xyz):
        self._node_xyz = node_xyz

    def draw_external_forces(self, gradient=True, scale=1.0):
        node_xyz = self.node_xyz

        volmesh = self.diagram.dual
        network = self.diagram

        load_vectors = {}
        load_mags = {}

        for cell in volmesh.cells_on_boundaries():
            vectors = []
            for face in volmesh.cell_faces(cell):
                if volmesh.is_halfface_on_boundary(face):
                    vectors.append(volmesh.halfface_normal(face, unitized=False))

            vectors = zip(*vectors)
            resultant = [sum(vector) for vector in vectors]
            load_vectors[cell] = resultant
            load_mags[cell] = length_vector(resultant)

        colordict = valuedict_to_colordict(load_mags, color_scheme=i_to_green)

        lines = []
        for vertex in load_vectors:
            vector = scale_vector(load_vectors[vertex], scale)
            sp = node_xyz[vertex]
            ep = add_vectors(sp, vector)
            if volmesh.attributes['convention'] == -1:
                sp, ep = ep, sp
            color = (0, 100, 0)
            if gradient:
                color = colordict[vertex]
            lines.append({
                'start': sp,
                'end': ep,
                'color': color,
                'arrow': 'end',
                'name': '{}.load.at.vertex.{}'.format(network.name, vertex)})

        return compas_rhino.draw_lines(lines,
                                       layer=self.layer,
                                       clear=False,
                                       redraw=False)

    def draw_pipes(self, edges, color, scale, tol):
        node_xyz = self.node_xyz
        cylinders = []

        forces = get_force_mags(self.diagram.dual, self.diagram)

        for edge in edges:
            u, v = edge
            start = node_xyz[u]
            end = node_xyz[v]
            force = abs(forces[edge])
            force = scale * force
            if force < tol:
                continue
            radius = sqrt(force / pi)
            if isinstance(color, dict):
                pipe_color = color[edge]
            else:
                pipe_color = color
            cylinders.append({
                'start': start,
                'end': end,
                'radius': radius,
                'color': pipe_color
            })
        return compas_rhino.draw_cylinders(cylinders, layer=self.layer, clear=False, redraw=False)
